# Protean
import pytest

from protean.core.exceptions import ObjectNotFoundError
from protean.core.queryset import Q

# Local/Relative Imports
from .elements import Person, PersonRepository, User


class TestDAODeleteFunctionality:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository, aggregate_cls=Person)
        test_domain.register(User)

    def test_delete_an_object_in_repository_by_id(self, test_domain):
        """ Delete an object in the reposoitory by ID"""
        person = test_domain.get_dao(Person).create(
            id=3, first_name="John", last_name="Doe", age=22
        )
        deleted_person = test_domain.get_dao(Person).delete(person)
        assert deleted_person is not None
        assert deleted_person.state_.is_destroyed is True

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(3)

    def test_delete_all_records_in_repository(self, test_domain):
        """Delete all objects in a repository"""

        test_domain.get_dao(Person).create(
            id=1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="dArtagnan", last_name="Musketeer", age=5
        )

        person_records = test_domain.get_dao(Person).query.filter(Q())
        assert person_records.total == 4

        test_domain.get_dao(Person).delete_all()

        person_records = test_domain.get_dao(Person).query.filter(Q())
        assert person_records.total == 0

    def test_deleting_a_persisted_entity(self, test_domain):
        """ Delete an object in the reposoitory by ID"""
        person = test_domain.get_dao(Person).create(
            id=3, first_name="Jim", last_name="Carrey"
        )
        deleted_person = test_domain.get_dao(Person).delete(person)
        assert deleted_person is not None
        assert deleted_person.state_.is_destroyed is True

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(3)

    def test_deleting_all_entities_of_a_type(self, test_domain):
        test_domain.get_dao(Person).create(
            id=1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="dArtagnan", last_name="Musketeer", age=5
        )

        people = test_domain.get_dao(Person).query.all()
        assert people.total == 4

        test_domain.get_dao(Person).delete_all()

        people = test_domain.get_dao(Person).query.all()
        assert people.total == 0

    def test_deleting_all_records_of_a_type_satisfying_a_filter(self, test_domain):
        test_domain.get_dao(Person).create(
            id=1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="d'Artagnan", last_name="Musketeer", age=5
        )

        # Perform update
        deleted_count = test_domain.get_dao(Person).query.filter(age__gt=3).delete_all()

        # Query and check if only the relevant records have been deleted
        assert deleted_count == 2

        person1 = test_domain.get_dao(Person).get(1)
        person2 = test_domain.get_dao(Person).get(2)

        assert person1 is not None
        assert person2 is not None

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(3)

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(4)

    def test_deleting_records_satisfying_a_filter(self, test_domain):
        test_domain.get_dao(Person).create(
            id=1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="d'Artagnan", last_name="Musketeer", age=5
        )

        # Perform update
        deleted_count = test_domain.get_dao(Person).query.filter(age__gt=3).delete()

        # Query and check if only the relevant records have been updated
        assert deleted_count == 2
        assert test_domain.get_dao(Person).query.all().total == 2

        assert test_domain.get_dao(Person).get(1) is not None
        assert test_domain.get_dao(Person).get(2) is not None
        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(3)

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(4)



# Standard Library Imports
from datetime import datetime

# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import DateTime, Dict, String
from protean.core.repository import BaseRepository
from protean.globals import current_domain


class EventLog(BaseAggregate):
    kind = String(max_length=50, required=True)
    payload = Dict(required=True)
    timestamp = DateTime(required=True, default=datetime.utcnow())


class EventLogRepository(BaseRepository):
    @classmethod
    def get_most_recent_event_by_type(cls, kind: str) -> EventLog:
        event_dao = current_domain.get_dao(EventLog)
        return (
            event_dao.query.filter(kind=kind.__name__)
            .order_by("-timestamp")
            .all()
            .first
        )

    @classmethod
    def get_all_events_of_type(cls, kind: str) -> EventLog:
        event_dao = current_domain.get_dao(EventLog)
        return (
            event_dao.query.filter(kind=kind.__name__)
            .order_by("-timestamp")
            .all()
            .items
        )

    class Meta:
        aggregate_cls = EventLog



# Protean
from protean.core.command import BaseCommand
from protean.core.field.basic import String


class UserRegistrationCommand(BaseCommand):
    email = String(required=True, max_length=250)
    username = String(required=True, max_length=50)
    password = String(required=True, max_length=255)



# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import String
from protean.utils import fully_qualified_name


class TestAggregateRegistration:
    def test_manual_registration_of_aggregate(self, test_domain):
        class User(BaseAggregate):
            name = String(max_length=50)

        test_domain.register(User)

        assert fully_qualified_name(User) in test_domain.registry.aggregates

    def test_settings_in_manual_registration(self, test_domain):
        class User(BaseAggregate):
            name = String(max_length=50)

            class Meta:
                provider = "foobar"
                model = "UserModel"

        test_domain.register(User)

        assert User.meta_.provider == "foobar"
        assert User.meta_.model == "UserModel"

    def test_setting_provider_in_decorator_based_registration(self, test_domain):
        @test_domain.aggregate(provider="foobar", model="UserModel")
        class User(BaseAggregate):
            name = String(max_length=50)

        test_domain.register(User)

        assert User.meta_.provider == "foobar"
        assert User.meta_.model == "UserModel"



# Protean
import pytest

from protean.core.field.basic import Integer, String
from protean.core.serializer import BaseSerializer
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import User, UserSchema


class TestSerializerInitialization:
    def test_that_base_serializer_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseSerializer()

    def test_that_a_concrete_serializer_can_be_instantiated(self):
        schema = UserSchema()
        assert schema is not None

    def test_that_meta_is_loaded_with_attributes(self):
        assert UserSchema.meta_.aggregate_cls is not None
        assert UserSchema.meta_.aggregate_cls == User

        assert UserSchema.meta_.declared_fields is not None
        assert all(key in UserSchema.meta_.declared_fields for key in ["name", "age"])


class TestSerializerRegistration:
    def test_that_serializer_can_be_registered_with_domain(self, test_domain):
        test_domain.register(UserSchema)

        assert fully_qualified_name(UserSchema) in test_domain.registry.serializers

    def test_that_serializer_can_be_registered_via_annotations(self, test_domain):
        @test_domain.serializer
        class PersonSchema:
            name = String(required=True)
            age = Integer(required=True)

            class Meta:
                aggregate_cls = User

        assert fully_qualified_name(PersonSchema) in test_domain.registry.serializers


class TestSerializerDump:
    def test_that_serializer_dumps_data_from_domain_element(self):
        user = User(name="John Doe", age=24)
        json_result = UserSchema().dump(user)
        assert json_result == {"age": 24, "name": "John Doe"}



"""
Custom Protean exception classes
"""
# Standard Library Imports
import logging

logger = logging.getLogger("protean")


class ProteanException(Exception):
    """Base class for all Exceptions raised within Protean"""

    def __init__(self, messages, **kwargs):
        logger.debug(f"Exception:: {messages}")
        self.messages = messages
        super().__init__(**kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}: {self.messages}"


class ConfigurationError(Exception):
    """Improper Configuration encountered like:
        * An important configuration variable is missing
        * Re-registration of Models
        * Incorrect associations
    """


class ObjectNotFoundError(ProteanException):
    """Object was not found, can raise 404"""


class TooManyObjectsError(Exception):
    """Expected one object, but found many"""


class InsufficientDataError(Exception):
    """Object was not supplied with sufficient data"""


class InvalidDataError(ProteanException):
    """Data (type, value) is invalid"""


class InvalidStateError(Exception):
    """Object is in invalid state for the given operation

    Equivalent to 409 (Conflict)"""


class InvalidOperationError(Exception):
    """Operation being performed is not permitted"""


class NotSupportedError(Exception):
    """Object does not support the operation being performed"""


class IncorrectUsageError(Exception):
    """Usage of a Domain Element violates principles"""


class ValidationError(ProteanException):
    """Raised when validation fails on a field. Validators and custom fields should
    raise this exception.

    :param errors: An error message or a list of error messages or a
    dictionary of error message where key is field name and value is error
    """


class UsecaseExecutionError(Exception):
    """ Raised when a failure response is encountered on executing a usecase

    :param value: a tuple comprising of the error code and error message
    :type value: tuple
    :param orig_exc: Optional original exception raised in the usecase
    :param orig_trace: Optional trace of the original exception in the usecase
    """

    def __init__(self, value, orig_exc=None, orig_trace=None, **kwargs):
        self.value = value
        self.orig_exc = orig_exc
        self.orig_trace = orig_trace

        super().__init__(**kwargs)



"""Module to test SQLAlchemy Provider Class"""
# Protean
import pytest

from protean.adapters.repository import Providers
from protean.adapters.repository.sqlalchemy import SAProvider
from sqlalchemy.engine.result import ResultProxy
from sqlalchemy.orm.session import Session

# Local/Relative Imports
from .elements import Alien, Person


@pytest.mark.postgresql
class TestProviders:
    """This class holds tests for Providers Singleton"""

    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(Alien)

    def test_initialization_of_providers_on_first_call(self, test_domain):
        """Test that ``providers`` object is available"""
        assert isinstance(test_domain.providers, Providers)
        assert test_domain.providers._providers is None

        test_domain.get_provider("default")
        assert test_domain.providers is not None

    def test_provider_detail(self, test_domain):
        """Test provider info loaded for tests"""

        provider1 = test_domain.get_provider("default")
        assert isinstance(provider1, SAProvider)

    def test_provider_get_connection(self, test_domain):
        """Test ``get_connection`` method and check for connection details"""

        conn = test_domain.get_provider("default").get_connection()
        assert conn is not None
        assert isinstance(conn, Session)

    def test_provider_raw(self, test_domain):
        """Test raw queries"""
        test_domain.get_dao(Person).create(
            first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(first_name="Jean", age=3, last_name="John")
        test_domain.get_dao(Person).create(first_name="Bart", age=6, last_name="Carrie")

        test_domain.get_dao(Alien).create(
            first_name="Sully", age=28, last_name="Monster"
        )
        test_domain.get_dao(Alien).create(
            first_name="Mike", age=26, last_name="Monster"
        )
        test_domain.get_dao(Alien).create(first_name="Boo", age=2, last_name="Human")

        provider = test_domain.get_provider("default")

        # Filter by column value
        results = provider.raw("SELECT count(*) FROM person where last_name = 'John'")
        assert isinstance(results, ResultProxy)
        assert next(results)[0] == 2

        results = provider.raw(
            "SELECT count(*) FROM person where last_name = 'John' and age = 3"
        )
        assert next(results)[0] == 1

        # This query brings results from multiple repositories
        results = provider.raw("SELECT count(*) FROM person where age in (6,3)")
        assert next(results)[0] == 2

        results = provider.raw(
            "SELECT * FROM person where last_name = 'John' and age in (6,7)"
        )
        assert next(results)[0] == "Murdock"



# Standard Library Imports
import logging

# Protean
from protean.utils import DomainObjects, derive_element_class
from protean.utils.container import BaseContainer

logger = logging.getLogger("protean.event")


class BaseDomainEvent(BaseContainer):
    """Base DomainEvent class that all other Domain Events should inherit from.

    Core functionality associated with Domain Events, like timestamping, are specified
    as part of the base DomainEvent class.
    """

    element_type = DomainObjects.DOMAIN_EVENT

    def __new__(cls, *args, **kwargs):
        if cls is BaseDomainEvent:
            raise TypeError("BaseDomainEvent cannot be instantiated")
        return super().__new__(cls)


def domain_event_factory(element_cls, **kwargs):
    return derive_element_class(element_cls, BaseDomainEvent)



# Adapters
# Protean
from protean.adapters.broker import Brokers
from protean.adapters.broker.celery import CeleryBroker, ProteanTask
from protean.adapters.broker.inline import InlineBroker
from protean.adapters.email import EmailProviders
from protean.adapters.email.dummy import DummyEmailProvider
from protean.adapters.email.sendgrid import SendgridEmailProvider
from protean.adapters.repository import Providers
from protean.adapters.repository.elasticsearch import ElasticsearchModel, ESProvider
from protean.adapters.repository.memory import MemoryModel, MemoryProvider
from protean.adapters.repository.sqlalchemy import SAProvider, SqlalchemyModel

__all__ = (
    "Brokers",
    "CeleryBroker",
    "ProteanTask",
    "InlineBroker",
    "EmailProviders",
    "DummyEmailProvider",
    "SendgridEmailProvider",
    "Providers",
    "ESProvider",
    "ElasticsearchModel",
    "SAProvider",
    "SqlalchemyModel",
    "MemoryProvider",
    "MemoryModel",
)



# Standard Library Imports
import os

# Protean
import pytest


def initialize_domain():
    # Protean
    from protean.domain import Domain

    domain = Domain("SQLAlchemy Test - SQLite")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    return domain


@pytest.fixture(autouse=True)
def test_domain():
    domain = initialize_domain()

    with domain.domain_context():
        yield domain


@pytest.fixture(scope="session", autouse=True)
def setup_db():
    domain = initialize_domain()
    with domain.domain_context():
        # Create all associated tables
        # Local/Relative Imports
        from .elements import Alien, ComplexUser, Person, User

        domain.register(Person)
        domain.register(Alien)
        domain.register(User)
        domain.register(ComplexUser)

        domain.get_dao(Person)
        domain.get_dao(Alien)
        domain.get_dao(User)
        domain.get_dao(ComplexUser)

        for provider in domain.providers_list():
            provider._metadata.create_all()

        yield

        # Drop all tables at the end of test suite
        for provider in domain.providers_list():
            provider._metadata.drop_all()


@pytest.fixture(autouse=True)
def run_around_tests(test_domain):
    yield
    if test_domain.providers.has_provider("default"):
        test_domain.get_provider("default")._data_reset()



# Local/Relative Imports
from .elements import Comment, Post


class TestReferenceField:
    def test_that_reference_field_has_a_shadow_attribute(self):
        assert "post_id" in Comment.meta_.attributes

    def test_that_reference_field_does_not_appear_among_fields(self):
        assert "post_id" not in Comment.meta_.declared_fields


class TestHasOneField:
    def test_that_has_one_field_appears_in_fields(self):
        assert "meta" in Post.meta_.declared_fields

    def test_that_has_one_field_does_not_appear_in_attributes(self):
        assert "meta" not in Post.meta_.attributes


class TestHasManyField:
    def test_that_has_many_field_appears_in_fields(self):
        assert "comments" in Post.meta_.declared_fields

    def test_that_has_many_field_does_not_appear_in_attributes(self):
        assert "comments" not in Post.meta_.attributes



# Protean
from protean.core.queryset import Q


class TestQ:
    """Class that holds tests for Q Objects"""

    def test_deconstruct(self):
        q = Q(price__gt=10.0)
        path, args, kwargs = q.deconstruct()
        assert path == "protean.utils.query.Q"
        assert args == ()
        assert kwargs == {"price__gt": 10.0}

    def test_deconstruct_negated(self):
        q = ~Q(price__gt=10.0)
        path, args, kwargs = q.deconstruct()
        assert args == ()
        assert kwargs == {
            "price__gt": 10.0,
            "_negated": True,
        }

    def test_deconstruct_or(self):
        q1 = Q(price__gt=10.0)
        q2 = Q(price=11.0)
        q3 = q1 | q2
        path, args, kwargs = q3.deconstruct()
        assert args == (("price__gt", 10.0), ("price", 11.0),)
        assert kwargs == {"_connector": "OR"}

    def test_deconstruct_and(self):
        q1 = Q(price__gt=10.0)
        q2 = Q(price=11.0)
        q = q1 & q2
        path, args, kwargs = q.deconstruct()
        assert args == (("price__gt", 10.0), ("price", 11.0),)
        assert kwargs == {}

    def test_deconstruct_multiple_kwargs(self):
        q = Q(price__gt=10.0, price=11.0)
        path, args, kwargs = q.deconstruct()
        assert args == (("price", 11.0), ("price__gt", 10.0),)
        assert kwargs == {}

    def test_reconstruct(self):
        q = Q(price__gt=10.0)
        path, args, kwargs = q.deconstruct()
        assert Q(*args, **kwargs) == q

    def test_reconstruct_negated(self):
        q = ~Q(price__gt=10.0)
        path, args, kwargs = q.deconstruct()
        assert Q(*args, **kwargs) == q

    def test_reconstruct_or(self):
        q1 = Q(price__gt=10.0)
        q2 = Q(price=11.0)
        q = q1 | q2
        path, args, kwargs = q.deconstruct()
        assert Q(*args, **kwargs) == q

    def test_reconstruct_and(self):
        q1 = Q(price__gt=10.0)
        q2 = Q(price=11.0)
        q = q1 & q2
        path, args, kwargs = q.deconstruct()
        assert Q(*args, **kwargs) == q



# Standard Library Imports
import os

# Protean
import pytest

from protean.domain import Domain
from protean.domain.config import Config

# config keys used for the TestConfig
TEST_KEY = "foo"
SECRET_KEY = "config"
non_key = "not-a-key"


def common_object_test(domain):
    assert domain.secret_key == "config"
    assert domain.config["TEST_KEY"] == "foo"
    assert "TestConfig" not in domain.config
    assert "non_key" not in domain.config


class TestConfig:
    def test_config_attribute_set(self):
        domain = Domain(__name__)
        domain.config.from_pyfile(__file__.rsplit(".", 1)[0] + ".py")
        domain.secret_key = "Baz"

        assert domain.secret_key == "Baz"
        assert domain.config["SECRET_KEY"] == "Baz"

    def test_config_repr(self):
        domain = Domain(__name__)
        domain.config.from_pyfile(__file__.rsplit(".", 1)[0] + ".py")

        assert repr(domain.config) == (
            "<Config {'ENV': 'production', 'DEBUG': False, 'SECRET_KEY': 'config', "
            "'AUTOLOAD_DOMAIN': True, "
            "'IDENTITY_STRATEGY': <IdentityStrategy.UUID: 1>, "
            "'IDENTITY_TYPE': <IdentityType.STRING: 'STRING'>, "
            "'DATABASES': {'default': {'PROVIDER': 'protean.adapters.MemoryProvider'}}, "
            "'BROKERS': {'default': {'PROVIDER': 'protean.adapters.InlineBroker'}}, "
            "'EMAIL_PROVIDERS': {'default': {'PROVIDER': 'protean.adapters.DummyEmailProvider', "
            "'DEFAULT_FROM_EMAIL': 'admin@team8solutions.com'}}, "
            "'AGGREGATE_CHILDREN_LIMIT': 100, "
            "'TEST_KEY': 'foo'}>"
        )

    def test_config_from_file(self):
        domain = Domain(__name__)
        domain.config.from_pyfile(__file__.rsplit(".", 1)[0] + ".py")
        common_object_test(domain)

    def test_config_from_object(self):
        domain = Domain(__name__)
        domain.config.from_object(__name__)
        common_object_test(domain)

    def test_config_from_json(self):
        domain = Domain(__name__)
        current_dir = os.path.dirname(os.path.abspath(__file__))
        domain.config.from_json(os.path.join(current_dir, "config.json"))
        common_object_test(domain)

    def test_config_from_mapping(self):
        domain = Domain(__name__)
        domain.config.from_mapping(
            {"SECRET_KEY": "config", "TEST_KEY": "foo", "non_key": "not-a-key"}
        )
        common_object_test(domain)

        domain = Domain(__name__)
        domain.config.from_mapping(
            [("SECRET_KEY", "config"), ("TEST_KEY", "foo"), ("non_key", "not-a-key")]
        )
        common_object_test(domain)

        domain = Domain(__name__)
        domain.config.from_mapping(
            SECRET_KEY="config", TEST_KEY="foo", non_key="not-a-key"
        )
        common_object_test(domain)

        domain = Domain(__name__)
        with pytest.raises(TypeError):
            domain.config.from_mapping({}, {})

    def test_config_from_class(self):
        class Base(object):
            TEST_KEY = "foo"

        class Test(Base):
            SECRET_KEY = "config"

        domain = Domain(__name__)
        domain.config.from_object(Test)
        common_object_test(domain)

    def test_config_from_envvar(self):
        env = os.environ
        try:
            os.environ = {}
            domain = Domain(__name__)
            with pytest.raises(RuntimeError) as e:
                domain.config.from_envvar("FOO_SETTINGS")
            assert "'FOO_SETTINGS' is not set" in str(e.value)
            assert not domain.config.from_envvar("FOO_SETTINGS", silent=True)

            os.environ = {"FOO_SETTINGS": __file__.rsplit(".", 1)[0] + ".py"}
            assert domain.config.from_envvar("FOO_SETTINGS")
            common_object_test(domain)
        finally:
            os.environ = env

    def test_config_from_envvar_missing(self):
        env = os.environ
        try:
            os.environ = {"FOO_SETTINGS": "missing.cfg"}
            with pytest.raises(IOError) as e:
                domain = Domain(__name__)
                domain.config.from_envvar("FOO_SETTINGS")
            msg = str(e.value)
            assert msg.startswith(
                "[Errno 2] Unable to load configuration "
                "file (No such file or directory):",
            )
            assert msg.endswith("missing.cfg'")
            assert not domain.config.from_envvar("FOO_SETTINGS", silent=True)
        finally:
            os.environ = env

    def test_config_missing(self):
        domain = Domain(__name__)
        with pytest.raises(IOError) as e:
            domain.config.from_pyfile("missing.cfg")
        msg = str(e.value)
        assert msg.startswith(
            "[Errno 2] Unable to load configuration "
            "file (No such file or directory):",
        )
        assert msg.endswith("missing.cfg'")
        assert not domain.config.from_pyfile("missing.cfg", silent=True)

    def test_config_missing_json(self):
        domain = Domain(__name__)
        with pytest.raises(IOError) as e:
            domain.config.from_json("missing.json")
        msg = str(e.value)
        assert msg.startswith(
            "[Errno 2] Unable to load configuration "
            "file (No such file or directory):",
        )
        assert msg.endswith("missing.json'")
        assert not domain.config.from_json("missing.json", silent=True)

    def test_custom_config_class(self):
        class SubConfig(Config):
            pass

        class SubDomain(Domain):
            config_class = SubConfig

        domain = SubDomain(__name__)
        assert isinstance(domain.config, SubConfig)
        domain.config.from_object(__name__)
        common_object_test(domain)

    def test_get_namespace(self):
        domain = Domain(__name__)
        domain.config["FOO_OPTION_1"] = "foo option 1"
        domain.config["FOO_OPTION_2"] = "foo option 2"
        domain.config["BAR_STUFF_1"] = "bar stuff 1"
        domain.config["BAR_STUFF_2"] = "bar stuff 2"
        foo_options = domain.config.get_namespace("FOO_")
        assert 2 == len(foo_options)
        assert "foo option 1" == foo_options["option_1"]
        assert "foo option 2" == foo_options["option_2"]
        bar_options = domain.config.get_namespace("BAR_", lowercase=False)
        assert 2 == len(bar_options)
        assert "bar stuff 1" == bar_options["STUFF_1"]
        assert "bar stuff 2" == bar_options["STUFF_2"]
        foo_options = domain.config.get_namespace("FOO_", trim_namespace=False)
        assert 2 == len(foo_options)
        assert "foo option 1" == foo_options["foo_option_1"]
        assert "foo option 2" == foo_options["foo_option_2"]
        bar_options = domain.config.get_namespace(
            "BAR_", lowercase=False, trim_namespace=False,
        )
        assert 2 == len(bar_options)
        assert "bar stuff 1" == bar_options["BAR_STUFF_1"]
        assert "bar stuff 2" == bar_options["BAR_STUFF_2"]



# Protean
import pytest

from protean.core.exceptions import ValidationError

# Local/Relative Imports
from .elements import Person, PersonRepository


class TestState:
    """Class that holds tests for Entity State Management"""

    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository, aggregate_cls=Person)

    def test_that_a_default_state_is_available_when_the_entity_instantiated(self):
        person = Person(first_name="John", last_name="Doe")
        assert person.state_ is not None
        assert person.state_._new
        assert person.state_.is_new
        assert not person.state_.is_persisted

    def test_that_retrieved_objects_are_not_marked_as_new(self, test_domain):
        person = test_domain.get_dao(Person).create(first_name="John", last_name="Doe")
        db_person = test_domain.get_dao(Person).get(person.id)

        assert not db_person.state_.is_new

    def test_that_entity_is_marked_as_saved_after_successful_persistence(
        self, test_domain
    ):
        person = Person(first_name="John", last_name="Doe")
        assert person.state_.is_new

        test_domain.get_dao(Person).save(person)
        assert person.state_.is_persisted

    def test_that_a_new_entity_still_shows_as_new_if_persistence_failed(
        self, test_domain
    ):
        person = Person(first_name="John", last_name="Doe")
        try:
            del person.first_name
            test_domain.get_dao(Person).save(person)
        except ValidationError:
            assert person.state_.is_new

    def test_that_a_changed_entity_still_shows_as_changed_if_persistence_failed(
        self, test_domain
    ):
        person = test_domain.get_dao(Person).create(first_name="John", last_name="Doe")

        person.first_name = "Jane"
        assert person.state_.is_changed

        try:
            del person.first_name
            test_domain.get_dao(Person).save(person)
        except ValidationError:
            assert person.state_.is_changed

    def test_that_entity_is_marked_as_not_new_after_successful_persistence(
        self, test_domain
    ):
        person = test_domain.get_dao(Person).create(first_name="John", last_name="Doe")
        assert not person.state_.is_new

    def test_that_aggregate_copying_resets_state_in_the_new_aggregate_object(
        self, test_domain
    ):
        person1 = test_domain.get_dao(Person).create(first_name="John", last_name="Doe")
        person2 = person1.clone()

        assert person2.state_.is_new

    def test_that_entity_marked_as_changed_if_attributes_are_updated(self, test_domain):
        person = test_domain.get_dao(Person).create(first_name="John", last_name="Doe")
        assert not person.state_.is_changed

        person.first_name = "Jane"
        assert person.state_.is_changed

    def test_that_entity_is_not_marked_as_changed_upon_attr_change_if_still_new(self):
        person = Person(first_name="John", last_name="Doe")
        assert not person.state_.is_changed

        person.first_name = "Jane Doe"
        assert not person.state_.is_changed

    def test_that_aggregate_is_marked_as_not_changed_after_save(self, test_domain):
        person = test_domain.get_dao(Person).create(first_name="John", last_name="Doe")
        person.first_name = "Jane"
        assert person.state_.is_changed

        test_domain.get_dao(Person).save(person)
        assert not person.state_.is_changed

    def test_that_an_entity_is_marked_as_destroyed_after_delete(self, test_domain):
        person = test_domain.get_dao(Person).create(first_name="John", last_name="Doe")
        assert not person.state_.is_destroyed

        test_domain.get_dao(Person).delete(person)
        assert person.state_.is_destroyed



"""Test Field Functionality"""

# Protean
import pytest

from protean.core.exceptions import ValidationError
from protean.core.field.basic import Field


class DummyStringField(Field):
    """This is a dummy Field class for testing"""

    default_error_messages = {
        "invalid_type": "Field value must be of str type.",
    }

    def _cast_to_type(self, value: str):
        """ Value must me a string type"""
        if type(value) != str:
            self.fail("invalid_type")
        return value


class MinLengthValidator:
    def __init__(self, min_length):
        self.min_length = min_length
        self.error = f"Ensure this value has at least " f"{self.min_length} character."

    def __call__(self, value):
        if len(value) < self.min_length:
            raise ValidationError(self.error)


class TestField:
    def test_init(self):
        """Test successful String Field initialization"""

        name = DummyStringField()
        assert name is not None

    def test_required(self):
        """Test errors if required field has no value"""

        with pytest.raises(ValidationError):
            name = DummyStringField(required=True)
            name._load(None)

    def test_defaults(self):
        """ Test default value is set when no value is supplied"""
        # Test with default value as constant
        name = DummyStringField(default="dummy")
        assert name._load("") == "dummy"

        # Test with default value as callable
        name = DummyStringField(default=lambda: "dummy")
        assert name._load("") == "dummy"

    def test_type_validation(self):
        """ Test type checking validation for the Field"""
        with pytest.raises(ValidationError):
            name = DummyStringField()
            name._load(1)

    def test_validators(self):
        """ Test custom validators defined for the field"""

        with pytest.raises(ValidationError):
            name = DummyStringField(validators=[MinLengthValidator(min_length=5)])
            name._load("Dum")

    def test_error_message(self):
        """ Test that proper error message is generated"""

        # Test the basic error message
        try:
            name = DummyStringField(required=True)
            name._load(None)
        except ValidationError as err:
            assert err.messages == {"unlinked": [name.error_messages["required"]]}

        # Test overriding of error message
        try:
            name = DummyStringField()
            name._load(1)
        except ValidationError as err:
            assert err.messages == {"unlinked": ["Field value must be of str type."]}

        # Test multiple error messages
        try:
            name = DummyStringField(
                validators=[
                    MinLengthValidator(min_length=5),
                    MinLengthValidator(min_length=5),
                ]
            )
            name._load("Dum")
        except ValidationError as err:
            assert err.messages == {
                "unlinked": [
                    "Ensure this value has at least 5 character.",
                    "Ensure this value has at least 5 character.",
                ]
            }

    def test_default_validators(self):
        """ Test that default validators for a Field are called"""

        def medium_string_validator(value):
            """Function checks the max length of a field"""
            if len(value) > 15:
                raise ValidationError("Value cannot be more than 15 characters long.")

        DummyStringField.default_validators = [medium_string_validator]
        with pytest.raises(ValidationError):
            name = DummyStringField()
            name._load("Dummy Dummy Dummy")



# Standard Library Imports
from uuid import UUID, uuid4

# Protean
import pytest

from protean.core.exceptions import InvalidOperationError, ValidationError
from protean.core.field.basic import Auto, String

# Local/Relative Imports
from .elements import Adult, NotAPerson, Person, PersonAutoSSN, PersonExplicitID


class TestProperties:
    def test_conversion_of_aggregate_values_to_dict(self):
        person = Person(id=12, first_name="John", last_name="Doe")
        assert person.to_dict() == {
            "id": 12,
            "first_name": "John",
            "last_name": "Doe",
            "age": 21,
        }

    def test_repr_output_of_aggregate(self):
        person = Person(id=12, first_name="John")

        assert str(person) == "Person object (id: 12)"
        assert repr(person) == "<Person: Person object (id: 12)>"


class TestIdentity:
    """Grouping of Identity related test cases"""

    def test_id_field_in_meta(self):
        assert hasattr(Person.meta_, "id_field")
        assert type(Person.meta_.id_field) is Auto

    def test_default_id_field_construction(self):
        assert "id" in Person.meta_.declared_fields
        assert "id" in Person.meta_.attributes

        assert type(Person.meta_.declared_fields["id"]) is Auto
        assert Person.meta_.id_field == Person.meta_.declared_fields["id"]

    def test_non_default_auto_id_field_construction(self):
        assert "id" not in PersonAutoSSN.meta_.declared_fields
        assert "id" not in PersonAutoSSN.meta_.attributes

        assert type(PersonAutoSSN.meta_.declared_fields["ssn"]) is Auto
        assert PersonAutoSSN.meta_.id_field.field_name == "ssn"
        assert (
            PersonAutoSSN.meta_.id_field == PersonAutoSSN.meta_.declared_fields["ssn"]
        )

    def test_non_default_explicit_id_field_construction(self, test_domain):
        assert "id" not in PersonExplicitID.meta_.declared_fields
        assert "id" not in PersonExplicitID.meta_.attributes

        assert type(PersonExplicitID.meta_.declared_fields["ssn"]) is String
        assert PersonExplicitID.meta_.id_field.field_name == "ssn"
        assert (
            PersonExplicitID.meta_.id_field
            == PersonExplicitID.meta_.declared_fields["ssn"]
        )


class TestIdentityValues:
    """Grouping of Identity value related test cases"""

    def test_default_id_field_value(self):
        person = Person(first_name="John", last_name="Doe")
        assert person.id is not None

        try:
            uuid_obj = UUID(str(person.id))
        except ValueError:
            pytest.fail("ID is not valid UUID")

        assert str(uuid_obj) == person.id

    def test_that_ids_are_immutable(self):
        """Test that `id` cannot be changed once assigned"""
        person = Person(first_name="John", last_name="Doe")

        with pytest.raises(InvalidOperationError):
            person.id = 13

    def test_non_default_auto_id_field_value(self):
        person = PersonAutoSSN(first_name="John", last_name="Doe")
        assert person.ssn is not None

        try:
            uuid_obj = UUID(str(person.ssn))
        except ValueError:
            pytest.fail("SSN is not valid UUID")

        assert str(uuid_obj) == person.ssn

    def test_non_default_explicit_id_field_value(self):
        with pytest.raises(ValidationError):
            PersonExplicitID(first_name="John Doe")

        new_uuid = uuid4()
        role = PersonExplicitID(ssn=new_uuid, first_name="John")
        assert role is not None
        assert role.ssn == str(new_uuid)

    def test_that_explicit_id_can_be_supplied_to_auto_id_field(self):
        new_uuid = uuid4()
        person = PersonExplicitID(ssn=new_uuid, first_name="John")
        assert person.ssn is not None
        assert person.ssn == str(new_uuid)


class TestEquivalence:
    def test_that_two_entities_with_same_id_are_treated_as_equal(self):
        person1 = Person(id=12345, first_name="John", last_name="Doe")
        person2 = Person(id=12346, first_name="John", last_name="Doe")

        assert person1 != person2  # Because their identities are different
        assert person2 != person1  # Because their identities are different

        person3 = Person(id=12345, first_name="John", last_name="Doe")
        assert (
            person1 == person3
        )  # Because it's the same record even though attributes differ
        assert person3 == person1

    def test_that_two_entities_of_different_types_are_different_even_with_same_id(self):
        """Test that two entities are not considered equal even if they have the same ID
            and one belongs to a different Entity class
        """
        not_a_person = NotAPerson(id=12345, first_name="John", last_name="Doe")
        person = Person(id=12345, first_name="John", last_name="Doe")

        assert not_a_person != person  # Even though their identities are the same
        assert person != not_a_person  # Even though their identities are the same

    def test_that_two_entities_of_inherited_types_are_different_even_with_same_id(self):
        """Test that two entities are not considered equal even if they have the same ID
            and one is subclassed from the other
        """
        adult = Adult(id=12345, first_name="John", last_name="Doe")
        person = Person(id=12345, first_name="John", last_name="Doe")

        assert adult != person  # Even though their identities are the same
        assert person != adult  # Even though their identities are the same

    def test_generated_aggregate_hash(self):
        """Test that the entity's hash is based on its identity"""
        hashed_id = hash(12345)

        person = Person(id=12345, first_name="John", last_name="Doe")
        assert hashed_id == hash(
            person
        )  # FIXME Should hash be based on ID alone, or other attrs too?

    def test_that_two_aggregates_that_are_equal_have_equal_hash(self):
        person1 = Person(id=12345, first_name="John", last_name="Doe")
        person2 = Person(id=12345, first_name="John", last_name="Doe")

        assert hash(person1) == hash(person2)



# Protean
import pytest

from protean.core.exceptions import ValidationError

# Local/Relative Imports
from .elements import Person, PersonRepository, User


class TestDAOSaveFunctionality:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository, aggregate_cls=Person)
        test_domain.register(User)

    def test_creation_throws_error_on_missing_fields(self, test_domain):
        """ Add an entity to the repository missing a required attribute"""
        with pytest.raises(ValidationError) as err:
            test_domain.get_dao(Person).create(last_name="Doe")

        assert err.value.messages == {"first_name": ["is required"]}

    def test_entity_persistence_with_create_method_and_fetch(self, test_domain):
        person = test_domain.get_dao(Person).create(
            id=11344234, first_name="John", last_name="Doe"
        )
        assert person is not None
        assert person.id == 11344234
        assert person.first_name == "John"
        assert person.last_name == "Doe"
        assert person.age == 21

        db_person = test_domain.get_dao(Person).get(11344234)
        assert db_person is not None
        assert db_person == person

    def test_multiple_persistence_for_an_aggregate(self, test_domain):
        """Test that save can be invoked again on an already existing entity, to update values"""
        person = Person(first_name="Johnny", last_name="John")
        test_domain.get_dao(Person).save(person)

        person.last_name = "Janey"
        test_domain.get_dao(Person).save(person)

        test_domain.get_dao(Person).get(person.id)
        assert person.last_name == "Janey"



"""Module for defining base Field class"""

# Standard Library Imports
import enum

from abc import ABCMeta, abstractmethod
from collections import defaultdict
from typing import Any, Iterable

# Protean
from protean.core import exceptions

# Local/Relative Imports
from .mixins import FieldDescriptorMixin

MISSING_ERROR_MESSAGE = (
    "ValidationError raised by `{class_name}`, but error key `{key}` does "
    "not exist in the `error_messages` dictionary."
)


class Field(FieldDescriptorMixin, metaclass=ABCMeta):
    """Abstract field from which other fields should extend.

    :param default: If set, this value will be used during entity loading
    if the field value is missing.
    :param required: if `True`, Raise a :exc:`ValidationError` if the field
    value is `None`.
    :param unique: Indicate if this field needs to be checked for uniqueness.
    :param label: Verbose name for this field
    :param choices: Valid choices for this field, if value is not one of the
    choices a `ValidationError` is raised.
    :param validators: Optional list of validators to be applied for this field.
    :param error_messages: Optional list of validators to be applied for
    this field.
    """

    # Default error messages for various kinds of errors.
    default_error_messages = {
        "invalid": "Value is not a valid type for this field.",
        "unique": "`{entity_name:s}` with this `{field_name:s}` already exists.",
        "required": "is required",
        "invalid_choice": "Value `{value!r}` is not a valid choice. "
        "Must be one of {choices!r}",
    }

    # Default validators for a Field
    default_validators = []

    # These values will trigger the self.required check.
    empty_values = (None, "", [], (), {})

    def __init__(
        self,
        referenced_as: str = None,
        identifier: bool = False,
        default: Any = None,
        required: bool = False,
        unique: bool = False,
        label: str = None,
        choices: enum.Enum = None,
        validators: Iterable = (),
        value=None,
        error_messages: dict = None,
    ):

        # Nothing to be passed into FieldCacheMixin for initialization
        super().__init__(referenced_as=referenced_as)

        self.identifier = identifier
        self.default = default

        # Indicates if field values need to be unique within the repository
        # always True for identifier field
        self.unique = True if self.identifier else unique

        # Indicates if this field is required, always True for identifier field
        self.required = True if self.identifier else required

        # Set the choices for this field
        self.choices = choices
        if self.choices:
            self.choice_dict = {}
            for _, member in self.choices.__members__.items():
                if isinstance(member.value, (tuple, list)):
                    self.choice_dict[member.value[0]] = member.value[1]
                else:
                    self.choice_dict[member.value] = member.value

        self.label = label
        self._validators = validators

        # Value holder
        self._value = value

        # Hold a reference to Entity registering the field
        self._entity_cls = None

        # Collect default error message from self and parent classes
        messages = {}
        for cls in reversed(self.__class__.__mro__):
            messages.update(getattr(cls, "default_error_messages", {}))
        messages.update(error_messages or {})
        self.error_messages = messages

    def __get__(self, instance, owner):
        if hasattr(instance, "__dict__"):
            return instance.__dict__.get(self.field_name, self.value)

    def __set__(self, instance, value):
        value = self._load(value)
        instance.__dict__[self.field_name] = value

        # Mark Entity as Dirty
        if hasattr(instance, "state_"):
            instance.state_.mark_changed()

    def __delete__(self, instance):
        instance.__dict__.pop(self.field_name, None)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value if value else None

    def fail(self, key, **kwargs):
        """A helper method that simply raises a `ValidationError`.
        """
        try:
            msg = self.error_messages[key]
        except KeyError:
            class_name = self.__class__.__name__
            msg = MISSING_ERROR_MESSAGE.format(class_name=class_name, key=key)
            raise AssertionError(msg)
        if isinstance(msg, str):
            msg = msg.format(**kwargs)

        # If a field is being used by itself (not owned by an entity/aggregate),
        #   it's field_name will be blank.
        field_name = self.field_name or "unlinked"
        raise exceptions.ValidationError({field_name: [msg]})

    @property
    def validators(self):
        """
        Some validators can't be created at field initialization time.
        This method provides a way to handle such default validators.
        """
        return [*self.default_validators, *self._validators]

    @abstractmethod
    def _cast_to_type(self, value: Any):
        """
        Abstract method to validate and convert the value passed to native type.
        All subclasses must implement this method.
        Raise a :exc:`ValidationError` if validation does not succeed.
        """
        pass

    def _run_validators(self, value):
        """Perform validation on ``value``. Raise a :exc:`ValidationError` if
        validation does not succeed.
        """
        if value in self.empty_values:
            return

        errors = defaultdict(list)
        for validator in self.validators:
            try:
                validator(value)
            except exceptions.ValidationError as err:
                field_name = self.field_name or "unlinked"
                errors[field_name].append(err.messages)

        if errors:
            raise exceptions.ValidationError(errors)

    def _load(self, value: Any):
        """
        Load the value for the field, run validators and return the value.
        Subclasses can override this to provide custom load logic.

        :param value: value of the field

        """

        # Check if value is one among recognized empty values, or is False (for complex objects)
        if value in self.empty_values:
            # If a default has been set for the field return it
            if self.default is not None:
                default = self.default
                value = default() if callable(default) else default
                return value

            # If no default is set and this field is required
            elif self.required:
                self.fail("required")

            # In all other cases just return the passed value, as we do not want to
            # run validations against an empty value
            # Because of this behavior, we preserve the data sanctity for int and float objects,
            # and return 0 or 0.0, as need be.
            elif value is None:
                return value

        # If choices exist then validate that value is be one of the choices
        if self.choices:
            value_list = value
            if not isinstance(value, (list, tuple)):
                value_list = [value]
            for v in value_list:
                if v not in self.choice_dict:
                    self.fail("invalid_choice", value=v, choices=list(self.choice_dict))

        # Cast and Validate the value for this Field
        value = self._cast_to_type(value)

        # Call the rest of the validators defined for this Field
        self._run_validators(value)

        return value

    def get_cache_name(self):
        return self.field_name



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mprotean` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``protean.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``protean.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
# Protean
import click


@click.group(invoke_without_command=True)
@click.version_option()
@click.pass_context
def main(ctx):
    """CLI utilities for the Protean"""
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@main.command()
def test():
    # Standard Library Imports
    import sys

    # Protean
    import pytest

    errno = pytest.main(["-vv", "--cache-clear", "--flake8"])

    sys.exit(errno)


@main.command()
def new():
    # Protean
    from cookiecutter.main import cookiecutter

    # Create project from the cookiecutter-protean.git repo template
    cookiecutter("gh:proteanhq/cookiecutter-protean")



# Protean
from protean.utils import Database, IdentityStrategy, IdentityType

DEBUG = True
TESTING = True
ENV = "development"

# A secret key for this particular Protean installation. Used in secret-key
# hashing algorithms.
SECRET_KEY = "tvTpk3PAfkGr5x9!2sFU%XpW7bR8cwKA"

# Database Configuration
DATABASES = {
    "default": {"PROVIDER": "protean.adapters.MemoryProvider"},
    "sqlite": {
        "PROVIDER": "protean.adapters.repository.sqlalchemy.SAProvider",
        "DATABASE": Database.SQLITE.value,
        "DATABASE_URI": "sqlite:///test.db",
    },
}

# Identity strategy to use when persisting Entities/Aggregates.
#
# Options:
#
#   * IdentityStrategy.UUID: Default option, and preferred. Identity is a UUID and generated during `build` time.
#       Persisted along with other details into the data store.
#   * IdentityStrategy.DATABASE: Let the database generate unique identity during persistence
#   * IdentityStrategy.FUNCTION: Special function that returns a unique identifier
IDENTITY_STRATEGY = IdentityStrategy.UUID

# Data type of Auto-Generated Identity Values
#
# Options:
#
#   * INTEGER
#   * STRING (Default)
IDENTITY_TYPE = IdentityType.STRING

# Messaging Mediums
BROKERS = {
    "default": {
        "PROVIDER": "protean.adapters.CeleryBroker",
        "URI": "redis://127.0.0.1:6379/2",
        "IS_ASYNC": True,
    },
}

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "console": {"format": "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "console",
        },
    },
    "loggers": {"protean": {"handlers": ["console"], "level": "DEBUG"}},
}



# Standard Library Imports
import logging

# Protean
from protean.core.exceptions import IncorrectUsageError, ValidationError
from protean.core.field.association import HasMany, HasOne
from protean.globals import current_domain
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger("protean.repository")


class _RepositoryMetaclass(type):
    """
    This base metaclass processes the class declaration and constructs a meta object that can
    be used to introspect the Repository class later. Specifically, it sets up a `meta_` attribute on
    the Repository to an instance of Meta, either the default of one that is defined in the
    Repository class.

    `meta_` is setup with these attributes:
        * `aggregate`: The aggregate associated with the repository
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Initialize Repository MetaClass and load attributes"""

        # Ensure initialization is only performed for subclasses of Repository
        # (excluding Repository class itself).
        parents = [b for b in bases if isinstance(b, _RepositoryMetaclass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Remove `abstract` in base classes if defined
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "abstract"):
                delattr(base.Meta, "abstract")

        new_class = super().__new__(mcs, name, bases, attrs, **kwargs)

        # Gather `Meta` class/object if defined
        attr_meta = attrs.pop("Meta", None)
        meta = attr_meta or getattr(new_class, "Meta", None)
        setattr(new_class, "meta_", RepositoryMeta(meta))

        return new_class


class RepositoryMeta:
    """ Metadata info for the Repository.

    An object of this class is either constructed automatically upon Repository class initialization, or
    a meta class can be explicitly passed along with the repository definition, as in inner `Meta` class.

    Options supported by Meta class:
    - ``aggregate_cls``: The aggregate associated with the repository
    """

    def __init__(self, meta):
        # The aggregate class with which the repository is associated
        #   Repository will be fetched via its associated aggregate.
        #
        # FIXME Validate that `aggregate_cls` is a Aggregate domain element
        self.aggregate_cls = getattr(meta, "aggregate_cls", None)


class BaseRepository(metaclass=_RepositoryMetaclass):
    """This is the baseclass for concrete Repository implementations.

    The three methods in this baseclass to `add`, `get` or `remove` entities are sufficient in most cases
    to handle application requirements. They have built-in support for handling child relationships and
    honor Unit of Work constructs. While they can be overriddent, it is generally suggested to call the
    parent method first before writing custom code.

    Repositories are strictly meant to be used in conjunction with Aggregate elements. While it is possible
    to link and use :ref:`entity`, it is always better to deal with persistence at the level of the transaction
    boundary, which is at the aggregate's level.
    """

    element_type = DomainObjects.REPOSITORY

    def __new__(cls, *args, **kwargs):
        # Prevent instantiation of `BaseRepository itself`
        if cls is BaseRepository:
            raise TypeError("BaseRepository cannot be instantiated")
        return super().__new__(cls)

    def add(self, aggregate):  # noqa: C901
        """This method helps persist or update aggregates into the persistence store.

        Returns the persisted aggregate.

        Protean adopts a collection-oriented design pattern to handle persistence. What this means is that
        the Repository interface does not hint in any way that there is an underlying persistence mechanism,
        avoiding any notion of saving or persisting data in the design layer. The task of syncing the data
        back into the persistence store is handled automatically.

        To be specific, a Repository mimics a `set` collection. Whatever the implementation, the repository
        will not allow instances of the same object to be added twice. Also, when retrieving objects from
        a Repository and modifying them, you don’t need to “re-save” them to the Repository.

        If there is a :ref:`Unit of Work <unit-of-work>` in progress, then the changes are performed on the
        UoW's active session. They are committed whenever the entire UoW is committed. If there is no
        transaction in progress, changes are committed immediately to the persistence store. This mechanism
        is part of the DAO's design, and is automatically used wherever one tries to persist data.
        """

        # Ensure that aggregate is clean and good to save
        # FIXME Let `clean()` raise validation errors
        errors = aggregate.clean() or {}
        # Raise any errors found during load
        if errors:
            logger.error(errors)
            raise ValidationError(errors)

        # If there are HasMany fields in the aggregate, sync child objects added/removed,
        #   but not yet persisted to the database.
        #
        # The details of in-transit child objects are maintained as part of the `has_many_field` itself
        #   in a variable called `_temp_cache`
        for field_name, field in aggregate.meta_.declared_fields.items():
            if isinstance(field, HasMany):
                has_many_field = getattr(aggregate, field_name)

                for item in has_many_field._temp_cache["removed"]:
                    dao = current_domain.get_dao(field.to_cls)
                    dao.delete(item)
                has_many_field._temp_cache[
                    "removed"
                ] = list()  # Empty contents of `removed` cache

                for item in has_many_field._temp_cache["added"]:
                    dao = current_domain.get_dao(field.to_cls)
                    item.state_.mark_new()
                    dao.save(item)
                has_many_field._temp_cache[
                    "added"
                ] = list()  # Empty contents of `added` cache

            if isinstance(field, HasOne):
                if field.has_changed:
                    dao = current_domain.get_dao(field.to_cls)
                    if field.change == "ADDED":
                        dao.save(field.value)
                    elif field.change == "UPDATED":
                        if field.change_old_value is not None:
                            # The object was replaced, so delete the old record
                            dao.delete(field.change_old_value)
                        else:
                            # The same object was updated, so mark it as new to be able to save
                            # FIXME This should have been automatic with `is_changed` flag in `state_`
                            field.value.state_.mark_new()

                        dao.save(field.value)
                    else:
                        dao.delete(field.change_old_value)

                    # Reset temporary fields after processing
                    field.change = None
                    field.change_old_value = None

        # Persist only if the aggregate object is new, or it has changed since last persistence
        if (not aggregate.state_.is_persisted) or (
            aggregate.state_.is_persisted and aggregate.state_.is_changed
        ):
            dao = current_domain.get_dao(self.meta_.aggregate_cls)
            dao.save(aggregate)

        return aggregate

    def remove(self, aggregate):
        """This method helps remove aggregates from the persistence store.

        Returns the removed aggregate.

        Protean mimics the behavior of a `set` collection in such methods. The repository promotes the
        illusion that we are dealing with collection like objects, so that the domain layer remains clean
        and oblivious to underlying persistence mechanisms. All changes are synced to the persistence store
        automatically by the Repository as and when appropriate.

        If there is a :ref:`Unit of Work <unit-of-work>` in progress, then the changes are performed on the
        UoW's active session. They are committed whenever the entire UoW is committed. If there is no
        transaction in progress, changes are committed immediately to the persistence store. This mechanism
        is part of the DAO's design, and is automatically used wherever one tries to persist data.
        """
        dao = current_domain.get_dao(self.meta_.aggregate_cls)
        dao.delete(aggregate)

        return aggregate

    def get(self, identifier):
        """This is a utility method to fetch data from the persistence store by its key identifier. All child objects,
        including enclosed entities, are returned as part of this call.

        Returns the fetched aggregate.

        All other data filtering capabilities can be implemented by using the underlying DAO's
        :meth:`BaseDAO.filter` method.

        Filter methods are typically implemented as domain-contextual queries, like `find_adults()`,
        `find_residents_of_area(zipcode)`, etc. It is also possible to make use of more complicated,
        domain-friendly design patterns like the `Specification` pattern.
        """
        dao = current_domain.get_dao(self.meta_.aggregate_cls)
        return dao.get(identifier)


def repository_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseRepository)

    if not (
        hasattr(element_cls.meta_, "aggregate_cls") and element_cls.meta_.aggregate_cls
    ):
        element_cls.meta_.aggregate_cls = kwargs.pop("aggregate_cls", None)

    if not element_cls.meta_.aggregate_cls:
        raise IncorrectUsageError(
            "Repositories need to be associated with an Aggregate"
        )

    return element_cls



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""Setup Module for Protean Application Framework"""

from __future__ import absolute_import, print_function

# Standard Library Imports
import io
import re

from glob import glob
from os.path import basename, dirname, join, splitext

# Protean
from setuptools import find_packages, setup


def read(*names, **kwargs):
    """Helper method to read files"""
    return io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8"),
    ).read()


elasticsearch_requires = ["elasticsearch-dsl>=7.1.0"]
redis_requires = ["redis==3.5.2"]
sqlite_requires = ["sqlalchemy>=1.3.15"]
postgresql_requires = ["psycopg2>=2.8.4", "sqlalchemy>=1.3.15"]
celery_requires = ["celery[redis]~=4.4.2"]
sendgrid_requires = ["sendgrid>=6.1.3"]
flask_requires = ["flask>=1.1.1"]
marshmallow_requires = ["marshmallow>=3.5.1"]

install_requires = marshmallow_requires + [
    "click>=7.0",
    "cookiecutter>=1.7.0",
    "inflection>=0.5.1",
    "python-dateutil>=2.8.1",
    "werkzeug>=1.0.0",
]

all_external_requires = [
    elasticsearch_requires
    + redis_requires
    + postgresql_requires
    + celery_requires
    + sendgrid_requires
    + flask_requires
    + marshmallow_requires
]

testing_requires = all_external_requires + [
    "mock==4.0.2",
    "passlib==1.7.2",
    "pluggy==0.13.1",
    "pyjwt==1.7.1",
    "pytest-cov==2.8.1",
    "pytest-flake8==1.0.6",
    "pytest-mock==3.1.0",
    "pytest-travis-fold==1.3.0",
    "pytest==5.4.2",
]

docs_requires = [
    "sphinx==3.0.3",
    "sphinx_rtd_theme==0.4.3",
]

dev_requires = (
    docs_requires
    + testing_requires
    + [
        "black==19.10b0",
        "check-manifest==0.42",
        "coverage==5.1",
        "docutils==0.16",
        "pre-commit==2.6.0",
        "tox==3.15.0",
        "twine==3.1.1",
    ]
)

setup(
    name="protean",
    version="0.5.6",
    license="BSD 3-Clause License",
    description="Protean Application Framework",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", read("README.rst")
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    author="Subhash Bhushan C",
    author_email="subhash@team8solutions.com",
    url="https://github.com/proteanhq/protean",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=["domain-driven design", "ddd", "cqrs", "cqs", "ports and adapters"],
    install_requires=install_requires,
    extras_require={
        "elasticsearch": elasticsearch_requires,
        "redis": redis_requires,
        "postgresql": postgresql_requires,
        "sqlite": sqlite_requires,
        "celery": celery_requires,
        "sendgrid": sendgrid_requires,
        "flask": flask_requires,
        "marshmallow": marshmallow_requires,
        "external": all_external_requires,
        "test": testing_requires,
        "tests": testing_requires,
        "testing": testing_requires,
        "dev": dev_requires,
        "docs": docs_requires,
        "all": dev_requires,
    },
    entry_points={"console_scripts": ["protean = protean.cli:main"]},
)



"""A Dummy Email Provider class"""
# Standard Library Imports
import logging

# Protean
from protean.core.email import BaseEmailProvider

logger = logging.getLogger("protean.adapters.email")


class DummyEmailProvider(BaseEmailProvider):
    """
    An email backend to simulate Email messages
    """

    def __init__(self, name, domain, conn_info):
        super().__init__(name, domain, conn_info)

    def send_email(self, message, dynamic_template=False):
        """Output message into log"""

        logger.debug("Email message dispatched: %s" % message.mime_message)
        return True



# Protean
import pytest

from protean.core.exceptions import ObjectNotFoundError

# Local/Relative Imports
from .elements import Person, PersonRepository, User


class TestDAOUpdateFunctionality:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository, aggregate_cls=Person)
        test_domain.register(User)

    def test_update_an_existing_entity_in_the_repository(self, test_domain):
        person = test_domain.get_dao(Person).create(
            id=11344234, first_name="John", last_name="Doe", age=22
        )

        test_domain.get_dao(Person).update(person, age=10)
        updated_person = test_domain.get_dao(Person).get(11344234)
        assert updated_person is not None
        assert updated_person.age == 10

    def test_that_updating_a_deleted_aggregate_raises_object_not_found_error(
        self, test_domain
    ):
        """Try to update a non-existing entry"""

        person = test_domain.get_dao(Person).create(
            id=11344234, first_name="Johnny", last_name="John"
        )
        test_domain.get_dao(Person).delete(person)
        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).update(person, {"age": 10})

    def test_updating_record_with_dictionary_args(self, test_domain):
        """ Update an existing entity in the repository"""
        person = test_domain.get_dao(Person).create(
            id=2, first_name="Johnny", last_name="John", age=2
        )

        test_domain.get_dao(Person).update(person, {"age": 10})
        u_person = test_domain.get_dao(Person).get(2)
        assert u_person is not None
        assert u_person.age == 10

    def test_updating_record_with_kwargs(self, test_domain):
        """ Update an existing entity in the repository"""
        person = test_domain.get_dao(Person).create(
            id=2, first_name="Johnny", last_name="John", age=2
        )

        test_domain.get_dao(Person).update(person, age=10)
        u_person = test_domain.get_dao(Person).get(2)
        assert u_person is not None
        assert u_person.age == 10

    def test_updating_record_with_both_dictionary_args_and_kwargs(self, test_domain):
        """ Update an existing entity in the repository"""
        person = test_domain.get_dao(Person).create(
            id=2, first_name="Johnny", last_name="John", age=2
        )

        test_domain.get_dao(Person).update(person, {"first_name": "Stephen"}, age=10)
        u_person = test_domain.get_dao(Person).get(2)
        assert u_person is not None
        assert u_person.age == 10
        assert u_person.first_name == "Stephen"

    def test_updating_record_through_filter(self, test_domain):
        """Test that update by query updates only correct records"""
        test_domain.get_dao(Person).create(
            id=1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="dArtagnan", last_name="Musketeer", age=5
        )

        # Perform update
        updated_count = (
            test_domain.get_dao(Person)
            .query.filter(age__gt=3)
            .update(last_name="Fraud")
        )

        # Query and check if only the relevant records have been updated
        assert updated_count == 2

        u_person1 = test_domain.get_dao(Person).get(1)
        u_person2 = test_domain.get_dao(Person).get(2)
        u_person3 = test_domain.get_dao(Person).get(3)
        u_person4 = test_domain.get_dao(Person).get(4)
        assert u_person1.last_name == "Musketeer"
        assert u_person2.last_name == "Musketeer"
        assert u_person3.last_name == "Fraud"
        assert u_person4.last_name == "Fraud"

    def test_updating_multiple_records_through_filter_with_arg_value(self, test_domain):
        """Try updating all records satisfying filter in one step, passing a dict"""
        test_domain.get_dao(Person).create(
            id=1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="dArtagnan", last_name="Musketeer", age=5
        )

        # Perform update
        updated_count = (
            test_domain.get_dao(Person)
            .query.filter(age__gt=3)
            .update_all({"last_name": "Fraud"})
        )

        # Query and check if only the relevant records have been updated
        assert updated_count == 2

        u_person1 = test_domain.get_dao(Person).get(1)
        u_person2 = test_domain.get_dao(Person).get(2)
        u_person3 = test_domain.get_dao(Person).get(3)
        u_person4 = test_domain.get_dao(Person).get(4)
        assert u_person1.last_name == "Musketeer"
        assert u_person2.last_name == "Musketeer"
        assert u_person3.last_name == "Fraud"
        assert u_person4.last_name == "Fraud"

    def test_updating_multiple_records_through_filter_with_kwarg_value(
        self, test_domain
    ):
        """Try updating all records satisfying filter in one step"""
        test_domain.get_dao(Person).create(
            id=1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="dArtagnan", last_name="Musketeer", age=5
        )

        # Perform update
        updated_count = (
            test_domain.get_dao(Person)
            .query.filter(age__gt=3)
            .update_all(last_name="Fraud")
        )

        # Query and check if only the relevant records have been updated
        assert updated_count == 2

        u_person1 = test_domain.get_dao(Person).get(1)
        u_person2 = test_domain.get_dao(Person).get(2)
        u_person3 = test_domain.get_dao(Person).get(3)
        u_person4 = test_domain.get_dao(Person).get(4)
        assert u_person1.last_name == "Musketeer"
        assert u_person2.last_name == "Musketeer"
        assert u_person3.last_name == "Fraud"
        assert u_person4.last_name == "Fraud"



# cSpell: disable

# Protean
from protean.utils import IdentityStrategy, IdentityType

DEBUG = True
TESTING = True
ENV = "development"

# A secret key for this particular Protean installation. Used in secret-key
# hashing algorithms.
SECRET_KEY = "tvTpk3PAfkGr5x9!2sFU%XpW7bR8cwKA"

# Database Configuration
DATABASES = {
    "default": {"PROVIDER": "protean.adapters.MemoryProvider"},
}

# Identity strategy to use when persisting Entities/Aggregates.
#
# Options:
#
#   * IdentityStrategy.UUID: Default option, and preferred. Identity is a UUID and generated during `build` time.
#       Persisted along with other details into the data store.
#   * IdentityStrategy.DATABASE: Let the database generate unique identity during persistence
#   * IdentityStrategy.FUNCTION: Special function that returns a unique identifier
IDENTITY_STRATEGY = IdentityStrategy.UUID

# Data type of Auto-Generated Identity Values
#
# Options:
#
#   * INTEGER
#   * STRING (Default)
IDENTITY_TYPE = IdentityType.STRING

# Messaging Mediums
BROKERS = {
    "default": {"PROVIDER": "protean.adapters.InlineBroker"},
}

EMAIL_PROVIDERS = {
    "default": {
        "PROVIDER": "protean.SendgridEmailProvider",
        "DEFAULT_FROM_EMAIL": "admin@team8solutions.com",
        "API_KEY": "this-is-a-fake-key",
    },
}



"""Utility module for Protean

Definitions/declaractions in this module should be independent of other modules,
to the maximum extent possible.
"""
# Standard Library Imports
import functools
import logging

from enum import Enum, auto

# Protean
from protean.core.exceptions import IncorrectUsageError

logger = logging.getLogger("protean.utils")


class IdentityStrategy(Enum):
    UUID = auto()
    DATABASE = auto()
    FUNCTION = auto()


class IdentityType(Enum):
    INTEGER = "INTEGER"
    STRING = "STRING"
    UUID = "UUID"


class Database(Enum):
    ELASTICSEARCH = "ELASTICSEARCH"
    MEMORY = "MEMORY"
    POSTGRESQL = "POSTGRESQL"
    SQLITE = "SQLITE"


def fully_qualified_name(cls):
    """Return Fully Qualified name along with module"""
    return ".".join([cls.__module__, cls.__name__])


def singleton(cls):
    """Make a class a Singleton class (only one instance)"""

    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance:
            wrapper_singleton.instance = cls(*args, **kwargs)
        return wrapper_singleton.instance

    wrapper_singleton.instance = None
    return wrapper_singleton


def convert_str_values_to_list(value):
    if not value:
        return []
    elif isinstance(value, str):
        return [value]
    else:
        return list(value)


class DomainObjects(Enum):
    AGGREGATE = "AGGREGATE"
    APPLICATION_SERVICE = "APPLICATION_SERVICE"
    COMMAND = "COMMAND"
    COMMAND_HANDLER = "COMMAND_HANDLER"
    DOMAIN_EVENT = "DOMAIN_EVENT"
    DOMAIN_SERVICE = "DOMAIN_SERVICE"
    EMAIL = "EMAIL"
    ENTITY = "ENTITY"
    MODEL = "MODEL"
    REPOSITORY = "REPOSITORY"
    SERIALIZER = "SERIALIZER"
    SUBSCRIBER = "SUBSCRIBER"
    VALUE_OBJECT = "VALUE_OBJECT"
    VIEW = "VIEW"


def derive_element_class(element_cls, base_cls):
    if not issubclass(element_cls, base_cls):
        try:
            new_dict = element_cls.__dict__.copy()
            new_dict.pop("__dict__", None)  # Remove __dict__ to prevent recursion

            element_cls = type(element_cls.__name__, (base_cls,), new_dict)
        except BaseException as exc:
            logger.debug("Error during Element registration:", repr(exc))
            raise IncorrectUsageError(
                "Invalid class {element_cls.__name__} for type {element_type.value}"
                " (Error: {exc})",
            )

    return element_cls



# Protean
import pytest

from celery import Task
from protean.adapters import CeleryBroker, ProteanTask
from protean.globals import current_domain
from tests.adapters.broker.celery_broker.elements import NotifySSOSubscriber, Person


class TestSubscriberNotifications:
    @pytest.fixture(autouse=True)
    def register(self):
        current_domain.register(Person)
        current_domain.register(NotifySSOSubscriber)

    @pytest.fixture
    def broker(self):
        return current_domain.brokers["default"]

    @pytest.fixture
    def decorated_task_obj(self, broker):
        return broker.construct_and_register_celery_task(NotifySSOSubscriber)

    def test_that_broker_is_celery(self, broker):
        assert isinstance(broker, CeleryBroker)

    def test_task_class_construction_and_registration(self, broker, decorated_task_obj):
        assert decorated_task_obj is not None
        assert isinstance(decorated_task_obj, ProteanTask)
        assert isinstance(decorated_task_obj, Task)

        assert (
            decorated_task_obj.name
            == "tests.adapters.broker.celery_broker.elements.notify_sso_subscriber"
        )
        assert decorated_task_obj.name in broker.celery_app.tasks

    @pytest.mark.skip(reason="Yet to implement")
    def test_queue_associated_with_subscriber(self):
        pass



# Standard Library Imports
from enum import Enum

# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import String
from protean.core.field.embedded import ValueObjectField
from protean.core.value_object import BaseValueObject


class FileType(Enum):
    PDF = "PDF"
    PPT = "PPT"


class File(BaseValueObject):
    url = String(max_length=1024)
    type = String(max_length=15, choices=FileType)

    class Meta:
        aggregate_cls = "Resource"


class Resource(BaseAggregate):
    title = String(required=True, max_length=50)
    associated_file = ValueObjectField(File)



# Standard Library Imports
import re

from collections import defaultdict
from typing import List

# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import Integer, String, Text
from protean.core.field.embedded import ValueObjectField
from protean.core.model import BaseModel
from protean.core.repository import BaseRepository
from protean.core.value_object import BaseValueObject
from protean.globals import current_domain


class Person(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)


class PersonRepository(BaseRepository):
    def find_adults(self, minimum_age: int = 21) -> List[Person]:
        return current_domain.get_dao(Person).filter(age__gte=minimum_age)

    class Meta:
        aggregate_cls = Person


class Email(BaseValueObject):
    REGEXP = r"\"?([-a-zA-Z0-9.`?{}]+@\w+\.\w+)\"?"

    # This is the external facing data attribute
    address = String(max_length=254, required=True)

    def clean(self):
        """ Business rules of Email address """
        errors = defaultdict(list)

        if not bool(re.match(Email.REGEXP, self.address)):
            errors["address"].append("is invalid")

        return errors


class User(BaseAggregate):
    email = ValueObjectField(Email, required=True)
    password = String(required=True, max_length=255)


class Provider(BaseAggregate):
    name = String()
    age = Integer()


class ProviderCustomModel(BaseModel):
    name = Text()

    class Meta:
        entity_cls = Provider
        schema_name = "adults"


class Receiver(BaseAggregate):
    name = String()
    age = Integer()



# Standard Library Imports
from enum import Enum

# Protean
from protean.core.field.basic import String
from protean.utils import DomainObjects


class DummyElement(Enum):
    FOO = "FOO"


class User:
    username = String(max_length=50)
    password = String(max_length=255)


class UserAggregate(User):
    element_type = DomainObjects.AGGREGATE


class UserEntity(User):
    element_type = DomainObjects.ENTITY


class UserVO(User):
    element_type = DomainObjects.VALUE_OBJECT


class UserFoo(User):
    element_type = DummyElement.FOO



# Standard Library Imports
from datetime import datetime, timedelta

# Protean
import pytest

from protean.core.exceptions import ObjectNotFoundError, TooManyObjectsError
from protean.core.queryset import QuerySet

# Local/Relative Imports
from .elements import Person, PersonRepository, User


class TestDAORetrievalFunctionality:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository, aggregate_cls=Person)
        test_domain.register(User)

    @pytest.fixture
    def persisted_person(self, test_domain):
        person = test_domain.get_dao(Person).create(
            id=1234, first_name="John", last_name="Doe"
        )
        return person

    def test_filtering_of_database_records(self, test_domain):
        # Add multiple entries to the database
        test_domain.get_dao(Person).create(first_name="John", last_name="Doe", age=22)
        test_domain.get_dao(Person).create(first_name="Jane", last_name="Doe", age=18)
        test_domain.get_dao(Person).create(first_name="Baby", last_name="Roe", age=2)

        # Filter by the last name
        people = test_domain.get_dao(Person).query.filter(last_name="Doe")
        assert people is not None
        assert people.total == 2
        assert len(people.items) == 2

        # Order the results by age
        people = test_domain.get_dao(Person).query.order_by("age")
        assert people is not None
        assert people.first.age == 2
        assert people.first.first_name == "Baby"

    def test_traversal_of_filter_results(self, test_domain):
        """ Test the traversal of the filter results"""
        for counter in range(1, 5):
            test_domain.get_dao(Person).create(
                id=counter, first_name=f"John{counter}", last_name="Doe"
            )

        people = test_domain.get_dao(Person).query.limit(2).order_by("id")
        assert people is not None
        assert people.total == 4
        assert len(people.items) == 2
        assert people.first.id == 1
        assert people.has_next
        assert not people.has_prev

        people = (
            test_domain.get_dao(Person).query.offset(2).limit(2).order_by("id").all()
        )
        assert len(people.items) == 2
        assert people.first.id == 3
        assert not people.has_next
        assert people.has_prev

    def test_entity_retrieval_by_its_primary_key(self, test_domain, persisted_person):
        """Test Entity Retrieval by its primary key"""
        dog = test_domain.get_dao(Person).get(persisted_person.id)
        assert dog is not None
        assert dog.id == 1234

    def test_failed_entity_retrieval_by_its_primary_key(self, test_domain):
        """Test failed Entity Retrieval by its primary key"""
        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(1235)

    def test_entity_retrieval_by_specific_column_value(
        self, test_domain, persisted_person
    ):
        dog = test_domain.get_dao(Person).find_by(first_name="John")
        assert dog is not None
        assert dog.id == 1234

    def test_failed_entity_retrieval_by_column_value(
        self, test_domain, persisted_person
    ):
        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).find_by(first_name="JohnnyChase")

    def test_entity_retrieval_by_multiple_column_values(self, test_domain):
        test_domain.get_dao(Person).create(
            id=2346, first_name="Johnny1", last_name="Bravo", age=8
        )
        test_domain.get_dao(Person).create(
            id=2347, first_name="Johnny2", last_name="Bravo", age=6
        )

        dog = test_domain.get_dao(Person).find_by(first_name="Johnny1", age=8)
        assert dog is not None
        assert dog.id == 2346

    def test_failed_entity_retrieval_by_multiple_column_values(self, test_domain):
        test_domain.get_dao(Person).create(
            id=2346, first_name="Johnny1", last_name="Bravo", age=8
        )
        test_domain.get_dao(Person).create(
            id=2347, first_name="Johnny2", last_name="Bravo", age=6
        )

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).find_by(first_name="Johnny1", age=6)

    def test_error_on_finding_multiple_results(self, test_domain):
        test_domain.get_dao(Person).create(
            id=2346, first_name="Johnny1", last_name="Bravo", age=8
        )
        test_domain.get_dao(Person).create(
            id=2347, first_name="Johnny1", last_name="Gravo", age=6
        )

        with pytest.raises(TooManyObjectsError):
            test_domain.get_dao(Person).find_by(first_name="Johnny1")

    def test_entity_query_initialization(self, test_domain):
        """Test the initialization of a QuerySet"""
        dao = test_domain.get_dao(Person)
        query = dao.query

        assert query is not None
        assert isinstance(query, QuerySet)
        assert query._criteria == QuerySet(dao, test_domain, Person)._criteria

    def test_filter_initialization_of_query_from_entity(self, test_domain):
        filters = [
            test_domain.get_dao(Person).query.filter(first_name="Murdock"),
            test_domain.get_dao(Person)
            .query.filter(first_name="Jean")
            .filter(last_name="John"),
            test_domain.get_dao(Person).query.offset(1),
            test_domain.get_dao(Person).query.limit(25),
            test_domain.get_dao(Person).query.order_by("first_name"),
            test_domain.get_dao(Person).query.exclude(first_name="Bravo"),
        ]

        for filter in filters:
            assert isinstance(filter, QuerySet)

    def test_that_chaining_filters_returns_a_queryset_for_further_chaining(
        self, test_domain
    ):
        person_query = test_domain.get_dao(Person).query.filter(name="Murdock")
        filters = [
            person_query,
            test_domain.get_dao(Person)
            .query.filter(first_name="Jean")
            .filter(last_name="John"),
            person_query.offset(5 * 5),
            person_query.limit(5),
            person_query.order_by("first_name"),
            person_query.exclude(last_name="Murdock"),
        ]

        for filter in filters:
            assert isinstance(filter, QuerySet)

    def test_filter_by_chaining_example_1(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Person attributes
        query = (
            test_domain.get_dao(Person)
            .query.filter(first_name="Jean")
            .filter(last_name="John")
            .filter(age=3)
        )
        people = query.all()

        assert people is not None
        assert people.total == 1
        assert len(people.items) == 1

        person = people.first
        assert person.id == 3

    def test_filter_by_chaining_example_2(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Person attributes
        query = test_domain.get_dao(Person).query.filter(last_name="John")
        people = query.all()

        assert people is not None
        assert people.total == 2
        assert len(people.items) == 2

        person = people.first
        assert person.id == 2

    def test_filter_by_chaining_example_3(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        people = query.all()

        assert people is not None
        assert people.total == 2
        assert len(people.items) == 2

        person = people.first
        assert person.id == 3

    def test_results_retrieved_with_filter(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by the LastName
        people = test_domain.get_dao(Person).query.filter(last_name="John")
        assert people is not None
        assert people.total == 2
        assert len(people.items) == 2

        # Order the results by age
        people = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("-age")
        )
        assert people is not None
        assert people.first.age == 7
        assert people.first.first_name == "Murdock"

    def test_results_retrieved_after_exclusion(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Exclusion
        dogs = test_domain.get_dao(Person).query.exclude(last_name="John")
        assert dogs is not None
        assert dogs.total == 1
        assert len(dogs.items) == 1
        assert dogs.first.age == 6
        assert dogs.first.first_name == "Bart"

    def test_results_retrieved_after_multiple_value_exclusion(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by the first_name
        people = test_domain.get_dao(Person).query.exclude(
            first_name__in=["Murdock", "Jean"]
        )
        assert people is not None
        assert people.total == 1
        assert len(people.items) == 1
        assert people.first.age == 6
        assert people.first.first_name == "Bart"

    def test_comparisons_using_different_operators(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="john"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by the Owner
        people_gte = test_domain.get_dao(Person).query.filter(age__gte=3)
        people_datetime_gte = test_domain.get_dao(Person).query.filter(
            created_at__gte=datetime.now() - timedelta(hours=24)
        )
        people_lte = test_domain.get_dao(Person).query.filter(age__lte=6)
        people_gt = test_domain.get_dao(Person).query.filter(age__gt=3)
        people_lt = test_domain.get_dao(Person).query.filter(age__lt=6)
        people_in = test_domain.get_dao(Person).query.filter(
            first_name__in=["Jean", "Bart", "Nobody"]
        )
        people_exact = test_domain.get_dao(Person).query.filter(last_name__exact="John")
        people_iexact = test_domain.get_dao(Person).query.filter(
            last_name__iexact="John"
        )
        people_contains = test_domain.get_dao(Person).query.filter(
            last_name__contains="Joh"
        )
        people_icontains = test_domain.get_dao(Person).query.filter(
            last_name__icontains="Joh"
        )

        assert people_gte.total == 3
        assert people_datetime_gte.total == 3
        assert people_lte.total == 2
        assert people_gt.total == 2
        assert people_lt.total == 1
        assert people_in.total == 2
        assert people_exact.total == 1
        assert people_iexact.total == 2
        assert people_contains.total == 1
        assert people_icontains.total == 2

    def test_exception_on_usage_of_unsupported_comparison_operator(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by the Owner
        with pytest.raises(NotImplementedError):
            test_domain.get_dao(Person).query.filter(age__notexact=3).all()



# Protean
import pytest

from mock import patch
from protean.adapters import DummyEmailProvider
from protean.core.email import BaseEmail
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import Person, PersonAdded, WelcomeEmail, WelcomeNewPerson


class TestEmailInitialization:
    def test_that_base_email_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseEmail()

    def test_that_email_can_be_instantiated(self):
        email = WelcomeEmail(to="john.doe@gmail.com", data={"foo": "bar"})
        assert email is not None


class TestEmailRegistration:
    def test_that_email_can_be_registered_with_domain(self, test_domain):
        test_domain.register(WelcomeEmail)

        assert fully_qualified_name(WelcomeEmail) in test_domain.registry.emails

    def test_that_email_can_be_registered_via_annotations(self, test_domain):
        @test_domain.email
        class AnnotatedEmail:
            def special_method(self):
                pass

        assert fully_qualified_name(AnnotatedEmail) in test_domain.registry.emails


class TestEmailTriggering:
    @patch.object(DummyEmailProvider, "send_email")
    def test_that_email_is_pushed_via_aggregate_command_method(self, mock, test_domain):
        test_domain.register(PersonAdded)
        test_domain.register(WelcomeEmail)
        test_domain.register(WelcomeNewPerson)

        Person.add_newcomer(
            {
                "email": "john.doe@gmail.com",
                "first_name": "John",
                "last_name": "Doe",
                "age": 21,
            }
        )
        mock.assert_called_once()



# Protean
import pytest

from protean.adapters.repository.sqlalchemy import Any, Contains, In, Overlap
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import List, String


class GenericPostgres(BaseAggregate):
    ids = List()
    role = String()


@pytest.mark.postgresql
class TestLookups:
    def test_any_lookup(self, test_domain):
        model_cls = test_domain.get_model(GenericPostgres)

        identifier = "foobar"
        lookup = Any("ids", identifier, model_cls)
        expr = lookup.as_expression()

        assert str(expr.compile()) == ":param_1 = ANY (public.generic_postgres.ids)"
        assert expr.compile().params == {"param_1": "foobar"}

    def test_contains_lookup_with_array(self, test_domain):
        model_cls = test_domain.get_model(GenericPostgres)

        identifier = ["foo", "bar"]
        lookup = Contains("ids", identifier, model_cls)
        expr = lookup.as_expression()

        assert str(expr.compile()) == "public.generic_postgres.ids @> :ids_1"
        assert expr.compile().params == {"ids_1": ["foo", "bar"]}

    def test_overlap_lookup_with_array(self, test_domain):
        model_cls = test_domain.get_model(GenericPostgres)

        identifier = ["foo", "bar"]
        lookup = Overlap("ids", identifier, model_cls)
        expr = lookup.as_expression()

        assert str(expr.compile()) == "public.generic_postgres.ids && :ids_1"
        assert expr.compile().params == {"ids_1": ["foo", "bar"]}

    def test_in_lookup(self, test_domain):
        model_cls = test_domain.get_model(GenericPostgres)

        target_roles = ["foo", "bar", "baz"]
        lookup = In("role", target_roles, model_cls)
        expr = lookup.as_expression()

        assert (
            str(expr.compile())
            == "public.generic_postgres.role IN (:role_1, :role_2, :role_3)"
        )
        assert expr.compile().params == {
            "role_1": "foo",
            "role_2": "bar",
            "role_3": "baz",
        }



""" Test cases for validators"""
# Standard Library Imports
import re

# Protean
import pytest

from protean.core.exceptions import ValidationError
from protean.core.field.validators import (
    MaxLengthValidator,
    MaxValueValidator,
    MinLengthValidator,
    MinValueValidator,
    RegexValidator,
)

TEST_DATA = [
    # (validator, value, expected),
    (MinLengthValidator(5), "abcde", None),
    (MinLengthValidator(5), "abcdef", None),
    (MinLengthValidator(5), "abcd", ValidationError),
    (MaxLengthValidator(10), "abcde", None),
    (MaxLengthValidator(10), "abcdefghij", None),
    (MaxLengthValidator(10), "abcdefghijkl", ValidationError),
    (MinValueValidator(100), 100, None),
    (MinValueValidator(100), 101, None),
    (MinValueValidator(100), 99, ValidationError),
    (MaxValueValidator(100), 100, None),
    (MaxValueValidator(100), 101, ValidationError),
    (MaxValueValidator(100), 99, None),
    (RegexValidator(), "", None),
    (RegexValidator(), "x1x2", None),
    (RegexValidator("[0-9]+"), "xxxxxx", ValidationError),
    (RegexValidator("[0-9]+"), "1234", None),
    (RegexValidator(re.compile("[0-9]+")), "1234", None),
    (RegexValidator(".*"), "", None),
    (RegexValidator(re.compile(".*")), "", None),
    (RegexValidator(".*"), "xxxxx", None),
    (RegexValidator("x"), "y", ValidationError),
    (RegexValidator(re.compile("x")), "y", ValidationError),
    (RegexValidator("x", inverse_match=True), "y", None),
    (RegexValidator(re.compile("x"), inverse_match=True), "y", None),
    (RegexValidator("x", inverse_match=True), "x", ValidationError),
    (RegexValidator(re.compile("x"), inverse_match=True), "x", ValidationError),
    (RegexValidator("x", flags=re.IGNORECASE), "y", ValidationError),
    (RegexValidator("a"), "A", ValidationError),
    (RegexValidator("a", flags=re.IGNORECASE), "A", None),
]


class TestValidators:
    def test_validators(self):
        for index, (validator, value, expected) in enumerate(TEST_DATA):
            exception_expected = expected is not None and issubclass(
                expected, Exception
            )

            print(
                "Test No: ",
                index,
                " - Validator: ",
                validator,
                " - value: ",
                value,
                " - expected: ",
                expected,
            )
            if exception_expected:
                with pytest.raises(expected):
                    validator(value)
            else:
                assert validator(value) is None



# Protean
import pytest

from protean.core.exceptions import InvalidOperationError, ValidationError

# Local/Relative Imports
from .elements import (
    Account,
    Balance,
    Currency,
    Email,
    MyOrgEmail,
    PolymorphicConnection,
    PolymorphicOwner,
    User,
)


class TestEquivalence:
    def test_two_value_objects_with_equal_values_are_considered_equal(self):
        email1 = Email.from_address("john.doe@gmail.com")
        email2 = Email.from_address("john.doe@gmail.com")

        assert email1 == email2

    def test_that_two_value_objects_with_different_values_are_different(self):
        email3 = Email.from_address("john.doe@gmail.com")
        email4 = Email.from_address("jane.doe@gmail.com")

        assert email3 != email4

    def test_that_two_value_objects_of_inherited_types_are_different_even_with_same_values(
        self,
    ):
        email = Email.from_address("john.doe@gmail.com")
        my_org_email = MyOrgEmail.from_address("john.doe@gmail.com")

        assert email != my_org_email

    def test_generated_value_object_hash(self):
        """Test that the entity's hash is based on its identity"""
        hashed_value = hash(frozenset({"address": "john.doe@gmail.com"}.items()))

        email = Email.from_address("john.doe@gmail.com")
        assert hash(email) == hashed_value

    def test_that_two_value_objects_that_are_equal_have_equal_hash(self):
        email1 = Email.from_address("john.doe@gmail.com")
        email2 = Email.from_address("john.doe@gmail.com")

        assert email1 == email2
        assert hash(email1) == hash(email2)


class TestEmailVOProperties:
    def test_output_to_dict(self):
        email = Email.from_address("john.doe@gmail.com")
        assert email.to_dict() == {"address": "john.doe@gmail.com"}

    def test_repr_output_of_value_object(self):
        email = Email.from_address("john.doe@gmail.com")
        assert (
            repr(email) == "<Email: Email object ({'address': 'john.doe@gmail.com'})>"
        )

    def test_str_output_of_value_object(self):
        email = Email.from_address("john.doe@gmail.com")
        assert str(email) == "Email object ({'address': 'john.doe@gmail.com'})"

    @pytest.mark.xfail
    def test_that_value_objects_are_immutable(self):
        email = Email.from_address(address="john.doe@gmail.com")
        with pytest.raises(InvalidOperationError):
            email.address = "jane.doe@gmail.com"


class TestEmailVOStructure:
    def test_email_vo_has_address_field(self):
        assert len(Email.meta_.declared_fields) == 1
        assert "address" in Email.meta_.declared_fields


class TestBalanceVOStructure:
    def test_balance_vo_has_currency_and_amount_fields(self):
        assert len(Balance.meta_.declared_fields) == 2
        assert "currency" in Balance.meta_.declared_fields
        assert "amount" in Balance.meta_.declared_fields

    def test_output_to_dict(self):
        balance = Balance.build(currency=Currency.USD.value, amount=0.0)
        assert balance.to_dict() == {"currency": "USD", "amount": 0.0}

    def test_repr_output_of_value_object(self):
        balance = Balance.build(currency=Currency.USD.value, amount=0.0)
        assert (
            repr(balance)
            == "<Balance: Balance object ({'currency': 'USD', 'amount': 0.0})>"
        )

    def test_str_output_of_value_object(self):
        balance = Balance.build(currency=Currency.USD.value, amount=0.0)
        assert str(balance) == "Balance object ({'currency': 'USD', 'amount': 0.0})"


class TestBalanceVOBehavior:
    def test_init(self):
        """Test that direct initialization works"""
        balance = Balance(currency=Currency.CAD.value, amount=0.0)
        assert balance is not None
        assert balance.currency == "CAD"
        assert balance.amount == 0.0

    def test_equivalence(self):
        """Test that two Balance VOs are equal if their values are equal"""
        balance1 = Balance.build(currency=Currency.USD.value, amount=0.0)
        balance2 = Balance.build(currency=Currency.USD.value, amount=0.0)

        assert balance1 == balance2

        balance3 = Balance.build(currency=Currency.INR.value, amount=0.0)

        assert balance3 != balance1

    def test_that_only_valid_currencies_can_be_assigned_to_balance_object(self):
        with pytest.raises(ValidationError):
            Balance.build(currency="FOO", amount=0.0)

    def test_that_only_valid_float_values_can_be_assigned_to_balance_object(self):
        with pytest.raises(ValidationError):
            Balance.build(currency="FOO", amount="abc")

    def test_that_a_negative_balance_less_than_one_trillion_is_invalid(self):
        with pytest.raises(ValidationError):
            Balance.build(currency="FOO", amount=-100000000000000.0)

    def test_that_new_balance_object_is_generated_with_replace_method(self):
        balance1 = Balance.build(currency=Currency.CAD.value, amount=0.0)
        balance2 = balance1.replace()

        assert balance2 is not balance1
        assert balance2 == balance1

        balance3 = balance1.replace(amount=150.0)
        assert balance1.amount == 0.0
        assert balance3.amount == 150.0

        balance4 = balance1.replace(currency="INR")
        assert balance1.currency == Currency.CAD.value
        assert balance4.currency == Currency.INR.value


class TestEmailVOBehavior:
    def test_validity(self):
        assert Email.validate("john.doe@gmail.com")
        assert not Email.validate("john.doe")
        assert not Email.validate("1234567890@gmail.com" * 26)

        with pytest.raises(ValueError):
            Email.from_address("john.doe")

    def test_init_from_constructor(self):
        email = Email(address="john.doe@gmail.com")
        assert email is not None
        assert email.address == "john.doe@gmail.com"

    def test_init_build(self):
        email = Email.from_address("john.doe@gmail.com")
        assert email is not None
        assert email.address == "john.doe@gmail.com"


class TestEmailVOEmbedding:
    def test_that_user_has_all_the_required_fields(self):
        assert all(
            field_name in User.meta_.declared_fields for field_name in ["email", "name"]
        )

    def test_that_user_can_be_initialized_successfully(self):
        user = User(email=Email.from_address("john.doe@gmail.com"))
        assert user is not None
        assert user.id is not None
        assert user.email == Email.from_address(address="john.doe@gmail.com")
        assert user.email_address == "john.doe@gmail.com"

    def test_that_user_can_be_initialized_successfully_with_embedded_fields(self):
        user = User(email_address="john.doe@gmail.com", name="John Doe")
        assert user is not None
        assert user.id is not None
        assert user.email == Email.from_address("john.doe@gmail.com")
        assert user.email_address == "john.doe@gmail.com"

    def test_that_mandatory_fields_are_validated(self):
        with pytest.raises(ValidationError) as multi_exceptions:
            User()

        assert "email_address" in multi_exceptions.value.messages
        assert multi_exceptions.value.messages["email_address"] == ["is required"]

        with pytest.raises(ValidationError) as email_exception:
            User(name="John Doe")

        assert "email_address" in email_exception.value.messages
        assert email_exception.value.messages["email_address"] == ["is required"]


class TestBalanceVOEmbedding:
    def test_that_account_has_all_the_required_fields(self):
        assert all(
            field_name in Account.meta_.declared_fields
            for field_name in ["balance", "kind"]
        )

    def test_that_account_can_be_initialized_successfully(self):
        account = Account(
            balance=Balance.build(currency="USD", amount=150.0), kind="PRIMARY"
        )
        assert account is not None
        assert account.id is not None
        assert account.balance == Balance.build(currency="USD", amount=150.0)
        assert account.balance_currency == "USD"
        assert account.balance_amount == 150.0

    def test_that_account_can_be_initialized_successfully_with_embedded_fields(self):
        account = Account(balance_currency="USD", balance_amount=150.0, kind="PRIMARY")
        assert account is not None
        assert account.id is not None
        assert account.balance == Balance.build(currency="USD", amount=150.0)
        assert account.balance_currency == "USD"
        assert account.balance_amount == 150.0

    def test_that_mandatory_fields_are_validated(self):
        with pytest.raises(ValidationError) as multi_exceptions:
            Account()

        assert "balance" in multi_exceptions.value.messages
        assert multi_exceptions.value.messages["balance"] == ["is required"]
        assert "kind" in multi_exceptions.value.messages
        assert multi_exceptions.value.messages["kind"] == ["is required"]

        with pytest.raises(ValidationError) as email_exception:
            Account(kind="PRIMARY")

        assert "balance" in email_exception.value.messages
        assert email_exception.value.messages["balance"] == ["is required"]


class TestNamedEmbedding:
    def test_that_explicit_names_are_used(self):
        assert len(PolymorphicConnection.meta_.declared_fields) == 2
        assert "connected_id" in PolymorphicConnection.meta_.declared_fields
        assert "connected_type" in PolymorphicConnection.meta_.declared_fields

    def test_that_explicit_names_are_preserved_in_aggregate(self):
        assert len(PolymorphicOwner.meta_.declared_fields) == 2
        assert "id" in PolymorphicOwner.meta_.declared_fields
        assert "connector" in PolymorphicOwner.meta_.declared_fields

        owner = PolymorphicOwner()
        assert owner.meta_.attributes is not None
        assert "connected_id" in owner.meta_.attributes
        assert "connected_type" in owner.meta_.attributes



# Protean
import pytest

from elasticsearch_dsl.query import Bool, Term

# Local/Relative Imports
from .elements import Person


@pytest.mark.elasticsearch
class TestESQuery:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)

    def test_construction_of_single_attribute_filter(self, test_domain):
        q1 = test_domain.get_dao(Person).query.filter(first_name="Jane")
        assert q1._owner_dao._build_filters(q1._criteria) == Term(first_name="Jane")

    def test_construction_of_filter_with_two_attributes(self, test_domain):
        q1 = test_domain.get_dao(Person).query.filter(
            first_name="Jane", last_name="Doe"
        )
        q2 = (
            test_domain.get_dao(Person)
            .query.filter(first_name="Jane")
            .filter(last_name="Doe")
        )

        filters1 = q1._owner_dao._build_filters(q1._criteria)
        filters2 = q2._owner_dao._build_filters(q2._criteria)

        assert filters1 == Bool(must=[Term(first_name="Jane"), Term(last_name="Doe")])
        assert filters2 == Bool(must=[Term(first_name="Jane"), Term(last_name="Doe")])

    def test_construction_of_single_attribute_negation_filter(self, test_domain):
        q1 = test_domain.get_dao(Person).query.exclude(first_name="Jane")
        assert q1._owner_dao._build_filters(q1._criteria) == Bool(
            must_not=[Term(first_name="Jane")]
        )

    def test_construction_of_filter_with_two_negated_attributes(self, test_domain):
        q1 = test_domain.get_dao(Person).query.exclude(
            first_name="Jane", last_name="Doe"
        )
        q2 = (
            test_domain.get_dao(Person)
            .query.exclude(first_name="Jane")
            .exclude(last_name="Doe")
        )

        filters1 = q1._owner_dao._build_filters(q1._criteria)
        filters2 = q2._owner_dao._build_filters(q2._criteria)

        assert filters1 == Bool(
            must_not=[Term(first_name="Jane"), Term(last_name="Doe")]
        )
        assert filters2 == Bool(
            must_not=[Term(first_name="Jane"), Term(last_name="Doe")]
        )

    def test_construction_with_combined_filter_and_exclude_with_filter_coming_first(
        self, test_domain
    ):
        q1 = test_domain.get_dao(Person).query.filter(last_name="Doe").exclude(age=3)
        filters1 = q1._owner_dao._build_filters(q1._criteria)
        assert filters1 == Bool(must=[Term(last_name="Doe")], must_not=[Term(age=3)])

    def test_construction_with_combined_filter_and_exclude_with_exclude_coming_first(
        self, test_domain
    ):
        q1 = test_domain.get_dao(Person).query.exclude(age=3).filter(last_name="Doe")
        filters1 = q1._owner_dao._build_filters(q1._criteria)
        assert filters1 == Bool(must=[Term(last_name="Doe")], must_not=[Term(age=3)])



# Protean
import pytest

from protean.adapters.broker.inline import InlineBroker
from protean.port.broker import BaseBroker


class TestBroker:
    def test_that_base_broker_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseBroker()

    def test_that_a_concrete_broker_can_be_initialized_successfully(self, test_domain):
        broker = InlineBroker("dummy_name", test_domain, {})

        assert broker is not None

    def test_that_domain_initializes_broker_from_config(self, test_domain):
        assert len(list(test_domain.brokers)) == 1
        assert isinstance(list(test_domain.brokers.values())[0], InlineBroker)



# Protean
from protean.core.exceptions import InvalidDataError, ValidationError
from protean.utils import DomainObjects, derive_element_class
from protean.utils.container import BaseContainer


class BaseCommand(BaseContainer):
    """The Base class for Protean-Compliant Domain Commands.

    Provides helper methods to custom define attributes, and find attribute names
    during runtime.

    Basic Usage::

        @domain.command
        class UserRegistrationCommand:
            email = String(required=True, max_length=250)
            username = String(required=True, max_length=50)
            password = String(required=True, max_length=255)


    (or)

        class UserRegistrationCommand(BaseCommand):
            email = String(required=True, max_length=250)
            username = String(required=True, max_length=50)
            password = String(required=True, max_length=255)

        domain.register_element(UserRegistrationCommand)
    """

    element_type = DomainObjects.COMMAND

    def __new__(cls, *args, **kwargs):
        if cls is BaseCommand:
            raise TypeError("BaseCommand cannot be instantiated")
        return super().__new__(cls)

    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except ValidationError as exception:
            raise InvalidDataError(exception.messages)


def command_factory(element_cls, **kwargs):
    return derive_element_class(element_cls, BaseCommand)



# Protean
from protean.core.entity import EntityMeta
from protean.core.field.basic import Auto, Integer, String

# Local/Relative Imports
from .elements import (
    AbstractPerson,
    Adult,
    AdultAbstractPerson,
    ConcretePerson,
    DbPerson,
    DifferentDbPerson,
    OrderedPerson,
    OrderedPersonSubclass,
    Person,
    SqlDifferentDbPerson,
    SqlPerson,
)


class TestEntityMeta:
    def test_entity_meta_structure(self):
        assert hasattr(Person, "meta_")
        assert type(Person.meta_) is EntityMeta

        # Persistence attributes
        # FIXME Should these be present as part of Entities, or a separate Model?
        assert hasattr(Person.meta_, "abstract")
        assert hasattr(Person.meta_, "schema_name")
        assert hasattr(Person.meta_, "provider")

        # Fields Meta Info
        assert hasattr(Person.meta_, "declared_fields")
        assert hasattr(Person.meta_, "attributes")
        assert hasattr(Person.meta_, "id_field")

        # Domain attributes
        assert hasattr(Person.meta_, "aggregate_cls")

    def test_entity_meta_has_declared_fields_on_construction(self):
        assert Person.meta_.declared_fields is not None
        assert all(
            key in Person.meta_.declared_fields.keys()
            for key in ["age", "first_name", "id", "last_name"]
        )

    def test_entity_declared_fields_hold_correct_field_types(self):
        assert type(Person.meta_.declared_fields["first_name"]) is String
        assert type(Person.meta_.declared_fields["last_name"]) is String
        assert type(Person.meta_.declared_fields["age"]) is Integer
        assert type(Person.meta_.declared_fields["id"]) is Auto

    def test_default_and_overridden_abstract_flag_in_meta(self):
        assert getattr(Person.meta_, "abstract") is False
        assert getattr(AbstractPerson.meta_, "abstract") is True

    def test_abstract_can_be_overridden_from_entity_abstract_class(self):
        """Test that `abstract` flag can be overridden"""

        assert hasattr(ConcretePerson.meta_, "abstract")
        assert getattr(ConcretePerson.meta_, "abstract") is False

    def test_abstract_can_be_overridden_from_entity_concrete_class(self):
        """Test that `abstract` flag can be overridden"""

        assert hasattr(AdultAbstractPerson.meta_, "abstract")
        assert getattr(AdultAbstractPerson.meta_, "abstract") is True

    def test_default_and_overridden_schema_name_in_meta(self):
        assert getattr(Person.meta_, "schema_name") == "person"
        assert getattr(DbPerson.meta_, "schema_name") == "pepes"

    def test_schema_name_can_be_overridden_in_entity_subclass(self):
        """Test that `schema_name` can be overridden"""
        assert hasattr(SqlPerson.meta_, "schema_name")
        assert getattr(SqlPerson.meta_, "schema_name") == "people"

    def test_default_and_overridden_provider_in_meta(self):
        assert getattr(Person.meta_, "provider") == "default"
        assert getattr(DifferentDbPerson.meta_, "provider") == "non-default"

    def test_provider_can_be_overridden_in_entity_subclass(self):
        """Test that `provider` can be overridden"""
        assert hasattr(SqlDifferentDbPerson.meta_, "provider")
        assert getattr(SqlDifferentDbPerson.meta_, "provider") == "non-default-sql"

    def test_default_and_overridden_order_by_in_meta(self):
        assert getattr(Person.meta_, "order_by") == ()
        assert getattr(OrderedPerson.meta_, "order_by") == ("first_name",)

    def test_order_by_can_be_overridden_in_entity_subclass(self):
        """Test that `order_by` can be overridden"""
        assert hasattr(OrderedPersonSubclass.meta_, "order_by")
        assert getattr(OrderedPersonSubclass.meta_, "order_by") == ("last_name",)

    def test_that_schema_is_not_inherited(self):
        assert Person.meta_.schema_name != Adult.meta_.schema_name



# Standard Library Imports
import random
import string

# Protean
import pytest

from protean.core.exceptions import ObjectNotFoundError
from protean.core.unit_of_work import UnitOfWork

# Local/Relative Imports
from .elements import Person, PersonRepository


@pytest.mark.postgresql
class TestTransactions:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository, aggregate_cls=Person)

    def random_name(self):
        return "".join(random.choices(string.ascii_uppercase + string.digits, k=15))

    def persisted_person(self, test_domain):
        repo = test_domain.repository_for(Person)
        person = Person(first_name=self.random_name(), last_name=self.random_name())
        repo.add(person)

        return person

    def test_new_objects_are_committed_as_part_of_one_transaction(self, test_domain):
        # Add a Person the database
        repo = test_domain.repository_for(Person)
        repo.add(self.persisted_person(test_domain))

        person_dao = test_domain.get_dao(Person)

        # Initiate a UnitOfWork Session
        with UnitOfWork():
            repo = test_domain.repository_for(Person)
            person2 = Person(first_name="Jane", last_name="Doe")
            repo.add(person2)

            # Test that the underlying database is untouched

            assert len(person_dao.outside_uow().query.all().items) == 1

        assert len(person_dao.query.all().items) == 2

    def test_updated_objects_are_committed_as_part_of_one_transaction(
        self, test_domain
    ):
        # Add a Person the database
        repo = test_domain.repository_for(Person)
        person = Person(first_name="John", last_name="Doe")
        repo.add(person)

        person_dao = test_domain.get_dao(Person)

        # Initiate a UnitOfWork Session
        with UnitOfWork():
            repo = test_domain.repository_for(Person)
            persisted_person = repo.get(person.id)

            persisted_person.last_name = "Dane"
            repo.add(persisted_person)

            # Test that the underlying database is untouched
            assert person_dao.outside_uow().find_by(id=person.id).last_name == "Doe"

        assert person_dao.get(person.id).last_name == "Dane"

    def test_deleted_objects_are_committed_as_part_of_one_transaction(
        self, test_domain
    ):
        # Add a Person the database
        repo = test_domain.repository_for(Person)
        person_to_be_added = self.persisted_person(test_domain)
        repo.add(person_to_be_added)

        person_dao = test_domain.get_dao(Person)

        # Initiate a UnitOfWork Session
        with UnitOfWork():
            repo = test_domain.repository_for(Person)
            persisted_person = repo.get(person_to_be_added.id)
            repo.remove(persisted_person)

            # Test that the underlying database is untouched
            assert len(person_dao.outside_uow().query.all().items) == 1

        assert len(person_dao.query.all().items) == 0

    def test_changed_objects_are_committed_as_part_of_one_transaction(
        self, test_domain
    ):
        # Add a Person the database
        repo = test_domain.repository_for(Person)
        person_to_be_updated = self.persisted_person(test_domain)
        person_to_be_deleted = self.persisted_person(test_domain)
        repo.add(person_to_be_updated)
        repo.add(person_to_be_deleted)

        person_dao = test_domain.get_dao(Person)

        # Initiate a UnitOfWork Session
        with UnitOfWork():
            repo_with_uow = test_domain.repository_for(Person)

            # Create a new person object to be added
            person_to_be_added = Person(first_name="John", last_name="Doe")
            repo_with_uow.add(person_to_be_added)

            # Update an existing Person record
            person_to_be_updated.last_name = "FooBar"
            repo_with_uow.add(person_to_be_updated)

            # Remove an existing Person record
            repo_with_uow.remove(person_to_be_deleted)

            # Test that the underlying database is untouched
            assert len(person_dao.query.all().items) == 2
            assert (
                person_dao.outside_uow().get(person_to_be_updated.id).last_name
                != "FooBar"
            )
            assert person_dao.get(person_to_be_deleted.id) is not None

        assert len(person_dao.query.all().items) == 2
        assert person_dao.get(person_to_be_added.id) is not None
        assert person_dao.get(person_to_be_updated.id).last_name == "FooBar"
        with pytest.raises(ObjectNotFoundError):
            person_dao.get(person_to_be_deleted.id)

    def test_changed_objects_are_committed_as_part_of_one_transaction_on_explict_commit(
        self, test_domain
    ):
        # Add a Person the database
        repo = test_domain.repository_for(Person)
        person_to_be_updated = self.persisted_person(test_domain)
        person_to_be_deleted = self.persisted_person(test_domain)
        repo.add(person_to_be_updated)
        repo.add(person_to_be_deleted)

        person_dao = test_domain.get_dao(Person)

        # Initiate a UnitOfWork Session
        uow = UnitOfWork()
        uow.start()

        repo_with_uow = test_domain.repository_for(Person)

        # Create a new person object to be added
        person_to_be_added = Person(first_name="John", last_name="Doe")
        repo_with_uow.add(person_to_be_added)

        # Update an existing Person record
        person_to_be_updated.last_name = "FooBar"
        repo_with_uow.add(person_to_be_updated)

        # Remove an existing Person record
        repo_with_uow.remove(person_to_be_deleted)

        # Test that the underlying database is untouched
        assert len(person_dao.query.all().items) == 2
        assert (
            person_dao.outside_uow().get(person_to_be_updated.id).last_name != "FooBar"
        )
        assert person_dao.get(person_to_be_deleted.id) is not None

        uow.commit()

        assert uow.in_progress is False
        assert len(person_dao.query.all().items) == 2
        assert person_dao.get(person_to_be_added.id) is not None
        assert person_dao.get(person_to_be_updated.id).last_name == "FooBar"
        with pytest.raises(ObjectNotFoundError):
            person_dao.get(person_to_be_deleted.id)

    def test_all_changes_are_discarded_on_rollback(self, test_domain):
        repo = test_domain.repository_for(Person)
        person_to_be_updated = self.persisted_person(test_domain)
        person_to_be_deleted = self.persisted_person(test_domain)
        repo.add(person_to_be_updated)
        repo.add(person_to_be_deleted)

        person_dao = test_domain.get_dao(Person)

        # Initiate a UnitOfWork Session
        uow = UnitOfWork()
        uow.start()

        repo_with_uow = test_domain.repository_for(Person)

        # Create a new person object to be added
        person_to_be_added = Person(first_name="John", last_name="Doe")
        repo_with_uow.add(person_to_be_added)

        # Update an existing Person record
        person_to_be_updated.last_name = "FooBar"
        repo_with_uow.add(person_to_be_updated)

        # Remove an existing Person record
        repo_with_uow.remove(person_to_be_deleted)

        # Test that the underlying database is untouched
        assert len(person_dao.query.all().items) == 2
        assert (
            person_dao.outside_uow().get(person_to_be_updated.id).last_name != "FooBar"
        )
        assert person_dao.get(person_to_be_deleted.id) is not None

        uow.rollback()

        assert uow.in_progress is False
        assert len(person_dao.query.all().items) == 2
        assert person_dao.get(person_to_be_updated.id).last_name != "FooBar"
        assert person_dao.get(person_to_be_deleted.id) is not None

    def test_session_is_destroyed_after_commit(self, test_domain):
        uow = UnitOfWork()
        uow.start()

        uow.commit()
        assert uow._sessions == {}
        assert uow.in_progress is False

    def test_session_is_destroyed_after_rollback(self, test_domain):
        uow = UnitOfWork()
        uow.start()

        uow.rollback()
        assert uow._sessions == {}
        assert uow.in_progress is False



# Protean
from protean.core.field.basic import String
from protean.core.view import BaseView
from protean.utils import fully_qualified_name


class TestViewRegistration:
    def test_manual_registration_of_view(self, test_domain):
        class Comment(BaseView):
            content = String(max_length=500)

        test_domain.register(Comment)

        assert fully_qualified_name(Comment) in test_domain.registry.views

    def test_setting_provider_in_decorator_based_registration(self, test_domain):
        @test_domain.view
        class Comment(BaseView):
            content = String(max_length=500)

        assert fully_qualified_name(Comment) in test_domain.registry.views



# Protean
import pytest

from protean.core.entity import BaseEntity
from protean.core.exceptions import ValidationError
from protean.core.field.basic import Boolean, Dict, Integer, List


class TestFields:
    @pytest.mark.xfail  # To be addressed as part of https://github.com/proteanhq/protean/issues/335
    def test_list_default(self):
        class Lottery(BaseEntity):
            numbers = List(content_type=Integer)

        lottery = Lottery()
        assert lottery.numbers is not None
        assert lottery.numbers == []

    def test_lists_can_be_mandatory(self):
        class Lottery(BaseEntity):
            jackpot = Boolean()
            numbers = List(content_type=Integer, required=True)

        with pytest.raises(ValidationError) as exc:
            Lottery(jackpot=True)

        assert exc.value.messages == {"numbers": ["is required"]}

    @pytest.mark.xfail  # To be addressed as part of https://github.com/proteanhq/protean/issues/335
    def test_dict_default(self):
        class Lottery(BaseEntity):
            numbers = Dict()

        lottery = Lottery()
        assert lottery.numbers is not None
        assert lottery.numbers == {}

    def test_dicts_can_be_mandatory(self):
        class Lottery(BaseEntity):
            jackpot = Boolean()
            numbers = Dict(required=True)

        with pytest.raises(ValidationError) as exc:
            Lottery(jackpot=True)

        assert exc.value.messages == {"numbers": ["is required"]}
